# functions for loading, manipulating and creating phylogenetic trees
from Bio import Phylo
import glob,os,shutil,subprocess
from multiprocessing import Pool
from .Tree import *
from xenoGI import genomes,fasta

#### Functions working with biopython trees

def prepareTree(bpTree,outGroupTaxaL):
    '''Given a biopython unrooted tree object (e.g. ASTRAL output) root,
name the internal nodes, and strip branch lengths (if
any). outGroupTaxaL provides the outgroups for rooting. Return
resulting tree in four tuple form.

    '''
    # strip confidence values in internal nodes (which mess up our
    # names)
    for n in bpTree.get_nonterminals():
        n.confidence = None
    
    bpTree = rootTree(bpTree,outGroupTaxaL)
    bpTree = nameInternalNodes(bpTree,"s")

    return bpTree

def rootTree(bpTree,outGroupTaxaL):
    '''Given a bio python tree, root it. outGroupTaxaL should consist of
all the species in the desired outgroup clade.'''

    bpTree.root_with_outgroup(*outGroupTaxaL)
    # rooting with outgroups with multiple species seems to leave a
    # trifurcating root.
    if len(bpTree.root)==2:
        return bpTree
    else:
        return makeRootBifurcation(bpTree,outGroupTaxaL)        

def makeRootBifurcation(bpTree,outGroupTaxaL):
    '''Given a bpTree that is not bifurcating at root, if it is
trifurcating convert to bifurcating. Otherwise thow error.'''

    if len(bpTree.root)!=3:
        raise ValueError("This tree has a strange number of branches at the root (not 2 or 3).")

    # find the two in outGroupTaxaL, and the one not in it
    inL=[]
    notIn=None
    for cld in bpTree.root:
        if tipsInOutgroupL(cld,outGroupTaxaL):
            inL.append(cld)
        else:
            notIn = cld

    if len(inL)!=2:
        raise ValueError("There should be two clades of this trifurcating tree which are in outGroupTaxaL, but there are not.")

    # We should take the branch length of notIn and split it up, one
    # half going to notIn, the other half to the new inL clade

    brToSplit=notIn.branch_length
    notIn.branch_length = brToSplit/2
    
    c1=Phylo.Newick.Clade(clades=inL,branch_length=brToSplit/2)
    c2=Phylo.Newick.Clade(clades=[c1,notIn],branch_length=0)
    outBpTree = Phylo.Newick.Tree(root=c2,rooted=True)

    return outBpTree
    
def tipsInOutgroupL(clade,outGroupTaxaL):
    '''Return True if all of the tips in clade are in outGroupTaxaL False
otherwise..'''
    for tipName in cladeTipNames(clade):
        if tipName not in outGroupTaxaL:
            return False
    return True
    
def cladeTipNames(clade):
    '''Get the names of all the tips in this clade.'''
    L=[]
    for tipC in clade.get_terminals():
        L.append(tipC.name)
    return L

def nameInternalNodes(bpTree,nameStem):
    '''Given a bio python tree without named internal nodes, name
them. This obliterates any names that might have been inside
already (e.g. if confidence values have been put in the name field).'''
    for j,n in enumerate(bpTree.get_nonterminals()):
        n.name=nameStem+str(j)
    return bpTree 

#### Functions for creating gene and species trees

def makeSpeciesTree(paramD,aabrhHardCoreL,genesO):
    '''Create a species tree based on the ortholog sets in aabrhHardCoreL.'''

    ## set up
    # create work dir if it doesn't already exist
    workDir = paramD['makeSpeciesTreeWorkingDir']
    if glob.glob(workDir)==[]:
        os.mkdir(workDir)

    numProcesses = paramD['numProcesses']
    javaPath = paramD['javaPath']
    astralPath = paramD['astralPath']
    astralTreeFN = paramD['astralTreeFN']
    gtFileStem = paramD['aabrhHardCoreGeneTreeFileStem']
    allGtFilePath = os.path.join(workDir,gtFileStem+'*.tre')
    aabrhHardCoreGeneTreesFN = paramD['aabrhHardCoreGeneTreesFN']
    outSpeciesTreeFN = paramD['speciesTreeFN'] # for main output
    outGroupTaxaL = [paramD['outGroup']]
    deleteSpeciesTreeWorkingDir = paramD['deleteSpeciesTreeWorkingDir']
    
    # if tree file already exists, throw error
    if os.path.isfile(outSpeciesTreeFN):
        raise IOError("The tree file " + outSpeciesTreeFN + " already exists.")

    # delete any pre-existing hard core gene trees
    for fn in glob.glob(allGtFilePath):
        os.remove(fn)
    
    ## make gene tree for each aabrh hard Core set
    # add numbering to list
    newAabrhHardCoreL = []
    for orthoNum,orthoT in enumerate(aabrhHardCoreL):
        newAabrhHardCoreL.append((orthoNum,orthoT))
    
    makeGeneTrees(paramD,True,genesO,workDir,gtFileStem,newAabrhHardCoreL)

    # remove alignments
    for fn in glob.glob(os.path.join(workDir,"align*.afa")):
        os.remove(fn)
    
    ## run Astral on gene trees

    # concatenate all gene tree files
    with open(aabrhHardCoreGeneTreesFN, 'w') as aabrhHardCoreGeneTreesF:
        for fn in sorted(glob.glob(allGtFilePath)):
            with open(fn) as infile:
                aabrhHardCoreGeneTreesF.write(infile.read())

    # run astral
    try:
        subprocess.check_call([javaPath, '-jar',astralPath, '-i', aabrhHardCoreGeneTreesFN, '-o',astralTreeFN],stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)
    except:
        raise Exception("Astral is giving an error on the file "+aabrhHardCoreGeneTreesFN)
        
    ## load and root the tree.
    speciesRtreeO = Rtree()
    speciesRtreeO.fromNewickFileLoadSpeciesTree(astralTreeFN,outGroupTaxaL)
    with open(outSpeciesTreeFN,"w") as f: # final output
        f.write(speciesRtreeO.toNewickStr()+";\n")

    ## delete the working directory
    if deleteSpeciesTreeWorkingDir:
        shutil.rmtree(workDir)
        
def makeGeneFamilyTrees(paramD,genesO,familiesO,gtFileStem = 'fam'):
    '''Given a families object, create a gene tree for each family.'''

    # create work dir if it doesn't already exist
    workDir = paramD['geneFamilyTreesDir']
    if glob.glob(workDir)==[]:
        os.mkdir(workDir)

    # create gene tuple for each family
    orthoTL = []
    for familyO in familiesO.iterFamilies():
        familyT = tuple(familyO.iterGenes())
        if len(familyT) > 1:
            # don't bother with single gene fams
            orthoTL.append((familyO.famNum,familyT))
            
    # make gene trees
    makeGeneTrees(paramD,False,genesO,workDir,gtFileStem,orthoTL)

    # remove alignments
    for fn in glob.glob(os.path.join(workDir,"align*.afa")):
        os.remove(fn)
    
    return

def makeGeneTrees(paramD,strainHeader,genesO,workDir,gtFileStem,orthoTL):
    '''Given a list of ortho groups, orthoTL, make a gene tree for each
and put in workDir. strainHeader is a boolean that if True means we
put the strain in the header for alignments (so we'll end up with that
for the names of the tips of the tree). gtFileStem gives the stem of
the name for the gene tree files. orthoTL is a list of
(orthoGroupNum,orthoT) where orthoT has the genes in an ortholog
group.

    '''
    numProcesses = paramD['numProcesses']    
    musclePath = paramD['musclePath']
    fastTreePath = paramD['fastTreePath']
    
    ## make gene trees
    argumentL = [([],[],paramD,strainHeader,genesO,workDir,gtFileStem,musclePath,fastTreePath) for i in range(numProcesses)]

    # set up argumentL
    counter = 0
    for orthoGroupNum,orthoT in orthoTL:
        # we use counter for placememnt in argumentL since at least in
        # the case of single gene families, some will be missing
        orthoGroupNumStr = str(orthoGroupNum).zfill(6) # pad w/ 0's so ls will display in right order
        argumentL[counter%numProcesses][0].append(orthoGroupNumStr)
        argumentL[counter%numProcesses][1].append(orthoT)
        counter += 1
        
    # run
    with Pool(processes=numProcesses) as p:
        for _ in p.imap_unordered(makeOneGeneTreeGroup, argumentL):
            pass

    return
        
def makeOneGeneTreeGroup(argT):
    '''Wrapper for multiprocessing. Given a group of orthoT's to work on,
calls makeOneGeneTree on each.'''
    
    orthoGroupNumStrL,orthoTL,paramD,strainHeader,genesO,workDir,gtFileStem,musclePath,fastTreePath = argT

    # get set of all genes in orthoTL to restrict size of seqD's
    orthoGenesS=set()
    for orthoT in orthoTL:
        orthoGenesS.update(orthoT)
    
    # load protein and dna sequences
    protSeqD=genomes.loadSeq(paramD, '_prot.fa',genesS=orthoGenesS)

    if paramD['dnaBasedGeneTrees'] == True:
        dnaSeqD = genomes.loadSeq(paramD, '_dna.fa',genesS=orthoGenesS)
    else:
        dnaSeqD = {}

    # make trees
    for i in range(len(orthoTL)):
        orthoGroupNumStr = orthoGroupNumStrL[i]
        orthoT = orthoTL[i]
        makeOneGeneTree(orthoGroupNumStr,orthoT,strainHeader,genesO,protSeqD,dnaSeqD,workDir,gtFileStem,musclePath,fastTreePath)
    
def makeOneGeneTree(orthoGroupNumStr,orthoT,strainHeader,genesO,protSeqD,dnaSeqD,workDir,gtFileStem,musclePath,fastTreePath):
    ''' Makes one gene tree from an ortho list. If dnaSeqD is empty, uses protein only.'''
 
    # get temp align file names
    inTempProtFN=os.path.join(workDir,"tempProt"+orthoGroupNumStr+".fa")
    outAlignFN=os.path.join(workDir,"align"+orthoGroupNumStr+".afa")

    # align
    alignOneOrthoT(orthoT,strainHeader,musclePath,inTempProtFN,outAlignFN,protSeqD,dnaSeqD,genesO)
    os.remove(inTempProtFN) # remove unaligned prot file
    
    # make gene tree
    geneTreeFN = os.path.join(workDir,gtFileStem+orthoGroupNumStr+".tre")
    if dnaSeqD == {}:
        # using protein
        subprocess.check_call([fastTreePath, '-out',geneTreeFN,outAlignFN],stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)
    else:
        # using dna
        subprocess.check_call([fastTreePath,'-gtr','-nt','-out',geneTreeFN,outAlignFN],stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)

def alignOneOrthoT(orthoT,strainHeader,musclePath,inProtFN,outAlignFN,protSeqD,dnaSeqD,genesO):
    '''Given genes in a single ortholog set, align them with muscle. If
dnaSeqD is empty, uses protein only.'''

    # write prots we want to align to temp file
    writeFasta(inProtFN,orthoT,strainHeader,genesO,protSeqD)

    # align proteins
    retCode = subprocess.call([musclePath, '-in' ,inProtFN, '-out', outAlignFN],stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)

    if retCode != 0:
        raise Exception("Alignment failed for "+inProtFN)
    
    if dnaSeqD != {}:
        # back align to get dna alignment, overwriting protein alignment.
        protAlignL = []
        for header,alignedProtSeq in fasta.load(outAlignFN):
            if strainHeader:
                protAlignL.append((int(header.rstrip().split()[1]),header,alignedProtSeq))
            else:
                protAlignL.append((int(header.rstrip()[1:]),header,alignedProtSeq))
        backAlign(outAlignFN,protAlignL,dnaSeqD,genesO)

    return
    
def writeFasta(inProtFN,orthoT,strainHeader,genesO,seqD):
    '''Writes a fasta block. seqs are specified in orthoT, and obtained
from seqD. If strainHeader is True, then we put strain whitespace gene
number in header. (In this case, the tree will end up with strain name
on its tips). Otherwise, only gene number.

    '''
    with open(inProtFN,"w") as f:
        for geneNum in orthoT:
            if strainHeader:
                strainName = genesO.numToStrainName(geneNum)
                f.write(">" + strainName + ' '  + str(geneNum) + "\n")
            else:
                f.write(">" + str(geneNum) + "\n")

            f.write(seqD[geneNum] + "\n")
            f.write("\n")

def backAlign(outAlignFN,protAlignL,dnaSeqD,genesO):
    '''Prints the nucleotide alignments given by one block of protein
    alignments.'''
    with open(outAlignFN,'w') as outf:
        printBlock=''
        for geneNum,header,protSeq in protAlignL:
            dnaSeq = dnaSeqD[geneNum]
            lenProtein = len(protSeq) - protSeq.count('-')
            if (lenProtein + 1) * 3 != len(dnaSeq):
                # dna has stop codon                          
                raise IndexError("Lengths of dna and protein do not correspond.")

            strainName = genesO.numToStrainName(geneNum)
            # add header
            printBlock += header + '\n'
            # adding gaps to the nucleotide sequence corresponding to gaps in 
            # the protein sequence and printing the nucleotide sequence
            printBlock += fixSeq(dnaSeq, protSeq) + "\n"
        print(printBlock,file=outf)

def fixSeq(sequence, protAlignment):
    """takes a nucleotide sequence and adds in gaps based on the corresponding
    protein alignment."""
    result = ""
    x = 0
    y = 0

    # while there are still nucleotides in the sequence
    while y != len(protAlignment):

        # if there is a gap in the protein alignment, add gaps to the nucleotide
        # sequence
        if protAlignment[y] == '-':
            result += "---"
            y += 1

        # if there is an amino acid, use the next codon
        else:
            result += sequence[3*x:3*(x+1)].upper()
            x += 1
            y += 1
    return result




# module for loading parameters.
import os

#### Base parameters

# These are parameters which users are not likely to need to change.
# We store them in a string of the same format as a parameters file.

baseParamStr = """

#### Genbank ####
# output from parsing genbank files

# gene order
geneOrderFN = 'geneOrder.txt'

# gene descriptions (for use in analysis)
geneInfoFN = 'geneInfo.txt'

# strain numbers and names
strainInfoFN = 'strainInfo.txt'

# unix style file path to fasta files
fastaFilePath = 'fasta/*.fa'

# listing of files with problems
problemGenbankFN = 'problemGenbankFiles.txt'


#### Blast ####

# unix style file path to blast output files
blastFilePath = 'blast/*.out'

# blast command line (except for value for evalue as well as db,query
# and outfiles)
blastCLine = 'blastp -matrix BLOSUM62 -gapopen 11 -gapextend 1 -seg yes -outfmt "6 qseqid sseqid evalue qlen qstart qend slen sstart send pident score" -evalue '

# Blast e-value threshold
evalueThresh = 1e-8

# alignCoverThresh is a threshold for the length of the blast alignment
# relative to query and subject length (ranges between 0 and 1)
alignCoverThresh = 0.65

# threshold for the percent identity in a hit
percIdentThresh = 0.35

# string to join the two strains compared in filename for blast
blastFileJoinStr = '_-VS-_'

#### Scores ####

# Note: for scores output files, if the extension we use here is
# .bout, the output will be saved in binary format. Otherwise it will
# be a less compact text based format
scoresFN = 'scores.bout'

# alignment parameters for making scores
# note, since parasail doesn't charge extend on the first base its
# like emboss, and not like ncbi blast. NCBI blast uses existance 11,
# extend 1 for blosum 62, thus we should use open 12 gapExtend 1 here.
gapOpen = 12
gapExtend = 1
matrix = 'parasail.blosum62'

# Synteny window size, that is the size of the neighborhood of each
# gene to consider when calculating synteny scores. (measured in
# number of genes). We go half this distance in either direction.
synWSize = 8

# When calculating synteny score between two genes, the number of
# pairs of scores to take (and average) from the neighborhoods of
# those two genes
numSynToTake = 3

# Core synteny window size, that is the number of core genes from the
# neighborhood of each gene to consider when calculating core synteny
# scores. (measured in number of genes). We go half this distance in
# each direction.
coreSynWsize = 20

# This helps determine the size of the arrays for our hash table of
# scores. It should be a value >= to 1. If one, it means the hash
# arrays have as many entries as there are score pairs.
hashArrayScaleFactor = 2

#### Making species trees ####

makeSpeciesTreeWorkingDir = 'makeSpeciesTreeWorkDir'
deleteSpeciesTreeWorkingDir = False # if True, we delete when done

# where to put gene trees for the aabrh hard core families
aabrhHardCoreGeneTreesFN = 'aabrhHardCoreGeneTrees.out'

# where to keep ASTRAL output. The final tree (which has been rooted,
# stripped of branch lengths and had internal nodes named) will be put
# in speciesTreeFN given above.
astralTreeFN = 'astralTree.tre'

#### Making gene trees ####

# directory to store trees for gene families
geneFamilyTreesDir = 'geneFamilyTrees'

# file stems
aabrhHardCoreGeneTreeFileStem = 'aabrhHardCoreFam'
blastFamGeneTreeFileStem = 'blastFam'


#### Family formation ####

# Hard core all around best reciprocal hit gene families
aabrhFN = 'aabrhHardCore.out'

# Other families
blastFamilyFN = 'blastFam.out'
initFamilyFN = 'initFam.out'
originFamilyFN = 'originFam.out'

# summary info about family formation
familyFormationSummaryFN = 'familyFormationSummary.out'

# Family formation involves several thresholds. We determine these by
# looking at histograms of scores.
scoreHistNumBins = 80

## Homology check
# parameters for finding the homologous (right) peak in raw score
# histograms. These put restrictions on the peaks we can
# find. widthRelHeight specifies where we measure the width of a peak,
# as a proportion of the way down from the top.
homologRightPeakLimit = 1.0
widthRelHeight = 0.9

# case 1 (normal case)
homologPeakWidthCase1 = 0.05
homologRequiredProminenceCase1 = 0.4
homologLeftPeakLimitCase1 = 0.65

# case 2 (extreme prominence. But allow to be very narrow)
homologPeakWidthCase2 = 0
homologRequiredProminenceCase2 = 6
homologLeftPeakLimitCase2 = 0.90

# case 3 (wide width with low prominence)
homologPeakWidthCase3 = 0.25
homologRequiredProminenceCase3 = 0.10
homologLeftPeakLimitCase3 = 0.65

# parameters for finding the non-homologous (left) peak in raw score
# histograms.
nonHomologPeakWidth = 0.15
nonHomologPeakProminence = 1
nonHomologLeftPeakLimit = 0
nonHomologRightPeakLimit = 0.6

## Splitting blast and initial families

# threshold of branch lengths for splitting unrooted gene trees from
# blast families. obtained by examining this distribution of maximum
# scores for aabrh hard core families.
quantileForObtainingSplitThresholds = .95
multiplierForObtainingSplitThresholds = 1.5

# The maximum size for blast and initial families is a multiple of the
# number of tips on the species tree. These parameters give the
# multiplier used. We want larger blast families than initial
# families, so the multiplier for blast should be larger.
maxBlastFamSizeMultiplier = 8
maxInitialFamSizeMultiplier = 4

# If some families still too large after splitting based on threshold,
# we force them to be split on a large internal branch. The choice of
# branch is based on branch length, and on achieving balance between
# the number of nodes on each side. This parameter affects how much
# weighting balance gets. Numbers > 1 mean we give more weight to
# balance.
forceSplitUtreeBalanceMultiplier = 10

## Synteny thresholds

# We determine synteny thresholds between pairs of strains based on
# the distribution of scores. quantileForObtainingSynThresholds gives
# the quantile we take from this dist. We then multiply this by
# multiplierForObtainingSynThresholds to get a threshold.
quantileForObtainingSynThresholds = 0.1
multiplierForObtainingSynThresholds = 0.5

#### LocusIsland formation ####

# island file
islandOutFN = 'islands.out'

# Summary info about island formation
islandFormationSummaryFN = 'islandFormationSummary.out'

# geneProximityRange tells us how far out we want to go from each gene
# in recording proximity for geneProximityD
geneProximityRange = 2

# In deciding whether to merge two islands, we judge partly based on
# the proximity of their genes. proximityThreshold defines what
# proximity is. A proximityThreshold of 1 means adjacent genes
proximityThresholdMerge = 1

# rscThreshold of 0 means that we merge islands if their rscore is 0
# or above
rscThresholdMerge = 0

# maxClusterSize is the maximum size that we make clusters of islands
# in the first step of the merging process
maxClusterSize = 50

#### Reconciliation ####

## reconciliation with costs permissive to origin events (for families
## that repeatedly insert in same syntenic location).

# File giving a list of xenoGI genes (one per line, string form) which
# we should use pemissive-origin reconciliation on. For each of these
# genes, we identify the initial family it belongs to, and then do
# reconciliation with permissive costs. By default, we set to None
# (and don't do this type of reconciliation). Users can override in
# the param.py file.
reconcilePermissiveOriginGeneListPath = None

# Costs to be used when we want to be permissive to origin events
DTLRcostPermissiveOrigin = 100
originCostPermissiveOrigin = 1

#### Family Refinement ####

# In family refinement, we consider alternate most parsimonious
# reconcilations. In some cases, the number of MPRs for a single
# family may be prohibitively large. upperNumMprThreshold specifies a
# limit beyond which we will randomly sample from the
# possibilities. This parameter also specifies the number of samples
# to be taken in that case.
upperNumMprThreshold = 20

# In order to test alternate MPRs, we must obtain nearby
# families. This parameter specifies how far to travel in each
# direction (in genes) in various genomes to collect those families.
geneProximityRangeRefineFamilies = 4

# In family refinement, we consider all origin families in islands
# this size or smaller
islandLenThresholdRefineFamilies = 2


#### Visualization and analysis ####

# analysis output
analysisDir = 'analysis'
islandsFNStem = 'islands'
genesFNstem = 'genes'

# output for browsers
bedFilePath = 'bed/*-island.bed' # unix style file path to bed output files

bedNumTries = 100 # number of random tries to find best coloring for islands

# The possible rgb values for bed files is based on the list below. 
potentialRgbL = ['245,130,48', '188,143,14','0,102,0','230,26,135','0,0,128', '145,30,180','0,255,255','128,0,0','0,255,0', '255,0,255','240,230,140','32,178,170','240,128,128','255,0,0','100,149,237','60,179,113','0,255,130','128,128,128','94,94,94','102,51,0']

"""


#### Functions
def createParametersD(baseParamStr,paramFN):
    '''Create and return a parameters dictionary. First parses the string
passed in as baseParamStr. This consists of parameters users are less
likely to modify. Then adds user specific parameters contained in the
file paramFN. Note that because the user parameters are put in the
parameters dictionary second, it is possible for a user to override
one of the base parameters simply by including that parameter in their
parameter file.

    '''
    paramD={}
    baseParamL = baseParamStr.split('\n')
    paramD = addParametersToD(baseParamL,paramD)

    f=open(paramFN,'r')
    userParamL = f.read().split('\n')
    paramD = addParametersToD(userParamL,paramD)
    f.close()
    
    return paramD
    
def addParametersToD(paramL,paramD):
    '''Given a list of lines (e.g. from a parameters file) add parameters
to paramD. Each line is a string. Some will be comments or blank
lines, and others will be python assignment statements. We use the
assignment statements to create entries in paramD.
    '''

    for s in paramL:
        if s == '' or s.lstrip() == '' or s.lstrip()[0]=='#':
            continue
        # it's not a blank line, a line with only whitespace or a
        # comment.  so it must be an assignment statement.
        key,value = s.rstrip().split('=')
        key = key.strip()
        value = value.strip()
        paramD[key] = eval(value)
    
    return paramD    
    
def loadFileNameMapD(fileNameMapFN,genbankFileList=None):
    '''Create a dictionary with mappings between genbank file names and
the human readable names we use in the tree. If fileNameMapFN contains
a string, we load that file and construct the mappings based on this.
Expects file with one species per line: genbank name + white space +
human name. If fileNameMapFN is None, we create the mappings between
the full file names (with path and extension) and the stem of the file
name, which in this case should correspond to what is in the input
tree.
    '''
    fileNameMapD = {}
    strainNamesL = []
    if fileNameMapFN == None:
        for fullPathFN in genbankFileList:
            fn = os.path.split(fullPathFN)[-1]
            stem = os.path.splitext(fn)[0]
            fileNameMapD[fn] = stem
            strainNamesL.append(stem)
    else:
        f = open(fileNameMapFN,'r')
        while True:
            s = f.readline()
            if s == '':
                break
            elif s[0].isspace():
                continue
            genbankStem,human = s.rstrip().split()
            fileNameMapD[genbankStem] = human
            strainNamesL.append(human)
    return fileNameMapD,tuple(strainNamesL)



"""Provides the entry point to xenoGI's functionality."""
__version__ = "3.0.0"
import sys, glob, os, readline, rlcompleter
from . import parameters,genbank,blast,trees,genomes,Score,scores,families,islands,analysis,islandBed
from .Tree import *
from .Family import *

def main():
    
    #### check command line
    try:
        assert(len(sys.argv) == 3)
        paramFN=sys.argv[1]
        task = sys.argv[2]
        assert(task in ['parseGenbank', 'runBlast', 'calcScores','makeSpeciesTree', 'makeFamilies', 'makeIslands','refine', 'printAnalysis', 'createIslandBed', 'plotScoreHists', 'interactiveAnalysis', 'runAll', 'version', 'debug'])
    
    except:
        print(
            """
   Exactly two arguments required.
      1. A path to a parameter file.
      2. The task to be run which must be one of: parseGenbank, runBlast, calcScores, makeSpeciesTree, makeFamilies, makeIslands, refine, printAnalysis, createIslandBed, plotScoreHists, interactiveAnalysis, runAll or version.

   For example: 
      xenoGI params.py parseGenbank
"""
            ,file=sys.stderr)
        sys.exit(1)

        
    #### load parameters and some other data we'll use below
    paramD = parameters.createParametersD(parameters.baseParamStr,paramFN)
        
    #### parseGenbank
    if task == 'parseGenbank':
        parseGenbankWrapper(paramD)
        
    #### runBlast
    elif task == 'runBlast':
        runBlastWrapper(paramD)
        
    #### calcScores
    elif task == 'calcScores':
        blastFnL=glob.glob(paramD['blastFilePath'])
        calcScoresWrapper(paramD,blastFnL)

    #### makeSpeciesTreeWrapper
    elif task == 'makeSpeciesTree':
        makeSpeciesTreeWrapper(paramD)
        
    #### makeFamilies
    elif task == 'makeFamilies':
        makeFamiliesWrapper(paramD)
        
    #### makeIslands
    elif task == 'makeIslands':
        makeIslandsWrapper(paramD)

    #### refineFamilies
    elif task == 'refine':
        refineWrapper(paramD)
        
    #### printAnalysis
    elif task == 'printAnalysis':
        printAnalysisWrapper(paramD,paramD['speciesTreeFN'],paramD['rootFocalClade'])

    #### createIslandBed
    elif task == 'createIslandBed':
        createIslandBedWrapper(paramD)

    #### interactiveAnalysis
    elif task == 'interactiveAnalysis':
        interactiveAnalysisWrapper(paramD)
        
    #### plotScoreHists
    elif task == 'plotScoreHists':
        scores.plotScoreHists(paramD)
        
    #### runAll
    elif task == 'runAll':
        parseGenbankWrapper(paramD)
        runBlastWrapper(paramD)
        blastFnL=glob.glob(paramD['blastFilePath'])
        calcScoresWrapper(paramD,blastFnL)
        makeFamiliesWrapper(paramD)
        makeIslandsWrapper(paramD)
        refineWrapper(paramD)
        printAnalysisWrapper(paramD,paramD['speciesTreeFN'],paramD['rootFocalClade'])
        createIslandBedWrapper(paramD)

    #### version
    elif task == 'version':
        print("xenoGI",__version__)
        
    #### debug
    elif task == 'debug':
        debugWrapper(paramD)

    #### validation via simulation
    elif task == 'simValidation':
        simValidationWrapper(paramD)
        
######## Task related functions

def parseGenbankWrapper(paramD):
    """Wrapper running stuff to parse genbank files."""
    genbankFileList=glob.glob(paramD['genbankFilePath'])

    # if directory for fastas doesn't exist yet, make it
    fastaDir = paramD['fastaFilePath'].split('*')[0]
    if glob.glob(fastaDir)==[]:
        os.mkdir(fastaDir)

    fileNameMapD,strainNamesT = parameters.loadFileNameMapD(paramD['fileNameMapFN'],genbankFileList)
    writeStrainInfoFN(strainNamesT,paramD)
    
    # parse
    genbank.parseGenbank(paramD,fastaDir,genbankFileList,fileNameMapD)

def writeStrainInfoFN(strainNamesT,paramD):
    """Write strain numbers and names to strainInfoFN. Numbers are given
by the index in strainNamesT."""
    with open(paramD['strainInfoFN'],'w') as f:
        for strainNum in range(len(strainNamesT)):
            f.write(str(strainNum)+"\t"+strainNamesT[strainNum]+"\n")

def readStrainInfoFN(strainInfoFN):
    """Read strainNamesT from strainInfoFN."""
    strainNamesL=[]
    with open(strainInfoFN,'r') as f:
        while True:
            s=f.readline()
            if s=='':
                break
            strainNum,strainName = s.rstrip().split("\t")
            strainNamesL.append(strainName)
    return tuple(strainNamesL)
    
def runBlastWrapper(paramD):
    """Wrapper to blast all genome files against each other."""
    strainNamesT = readStrainInfoFN(paramD['strainInfoFN'])
    dbFileL=blast.getDbFileL(paramD['fastaFilePath'],strainNamesT)
    blast.runBlast(dbFileL,dbFileL,paramD)
    
def loadGenomeRelatedData(paramD):
    """Load some data related to genomes and strains."""
    strainNamesT = readStrainInfoFN(paramD['strainInfoFN'])
    genesO = genomes.genes(paramD['geneInfoFN'])
    geneOrderD=genomes.createGeneOrderD(paramD['geneOrderFN'],None)
    return strainNamesT,genesO,geneOrderD

def calcScoresWrapper(paramD,blastFnL):
    """Wrapper running stuff to calculate scores."""

    strainNamesT,genesO,geneOrderD = loadGenomeRelatedData(paramD)
    
    # object for storing scores
    scoresO=Score.Score()
    scoresO.initializeDataAttributes(blastFnL,paramD,strainNamesT)

    ## similarity scores
    scoresO = scores.calcRawScores(paramD,scoresO)

    ## synteny scores
    scoresO = scores.calcSynScores(scoresO,geneOrderD,paramD)

    ## core synteny scores
    scoresO = scores.calcCoreSynScores(scoresO,strainNamesT,paramD,geneOrderD)

    # write scores to file
    scores.writeScores(scoresO,strainNamesT,paramD['scoresFN'])
    
def makeSpeciesTreeWrapper(paramD):
    '''call makeTree to create the species tree '''

    # need genesO should initializeGeneInfoD
    strainNamesT,genesO,geneOrderD = loadGenomeRelatedData(paramD)
    aabrhHardCoreL = scores.loadOrthos(paramD['aabrhFN'])
    trees.makeSpeciesTree(paramD,aabrhHardCoreL,genesO)

def loadTreeRelatedData(speciesTreeFN):
    """Load some data related to trees."""
    speciesRtreeO = Rtree()
    speciesRtreeO.fromNewickFileLoadSpeciesTree(speciesTreeFN)
    subtreeD=speciesRtreeO.createSubtreeD()
    return speciesRtreeO,subtreeD

def makeFamiliesWrapper(paramD):
    """Wrapper to create gene families."""

    strainNamesT,genesO,geneOrderD = loadGenomeRelatedData(paramD)
    speciesRtreeO,subtreeD = loadTreeRelatedData(paramD['speciesTreeFN'])

    ## read scores
    scoresO = scores.readScores(strainNamesT,paramD['scoresFN'])
    aabrhHardCoreL = scores.loadOrthos(paramD['aabrhFN'])

    ## make gene families
    with open(paramD['familyFormationSummaryFN'],'w') as familyFormationSummaryF:
        initialFamiliesO,originFamiliesO = families.createFamiliesO(speciesRtreeO,strainNamesT,scoresO,genesO,aabrhHardCoreL,paramD,familyFormationSummaryF)
        
def makeIslandsWrapper(paramD):
    """Wrapper to create islands"""

    strainNamesT,genesO,geneOrderD = loadGenomeRelatedData(paramD)
    speciesRtreeO,subtreeD = loadTreeRelatedData(paramD['speciesTreeFN'])
    
    ## read gene families
    originFamiliesO = families.readFamilies(paramD['originFamilyFN'],speciesRtreeO,genesO,"origin")

    ## group gene families into islands
    with open(paramD['islandFormationSummaryFN'],'w') as islandFormationSummaryF:
        locIslByNodeD = islands.makeLocusIslands(geneOrderD,subtreeD,speciesRtreeO,paramD,originFamiliesO,paramD['rootFocalClade'],islandFormationSummaryF)

def refineWrapper(paramD):
    """Wrapper to refine families and islands by considering alternate reconcilitions."""

    strainNamesT,genesO,geneOrderD = loadGenomeRelatedData(paramD)
    speciesRtreeO,subtreeD = loadTreeRelatedData(paramD['speciesTreeFN'])
    
    ## read gene families
    initialFamiliesO = families.readFamilies(paramD['initFamilyFN'],speciesRtreeO,genesO,"initial")
    originFamiliesO = families.readFamilies(paramD['originFamilyFN'],speciesRtreeO,genesO,"origin")
    islandByNodeD=islands.readIslands(paramD['islandOutFN'],speciesRtreeO)

    ## refine families
    with open(paramD['familyFormationSummaryFN'],'a') as familyFormationSummaryF:
        initialFamiliesO,originFamiliesO = families.refineFamilies(paramD,islandByNodeD,initialFamiliesO,originFamiliesO,geneOrderD,genesO,familyFormationSummaryF,strainNamesT)

    ## redo islands with new originFamilies
    with open(paramD['islandFormationSummaryFN'],'w') as islandFormationSummaryF:
        locIslByNodeD = islands.makeLocusIslands(geneOrderD,subtreeD,speciesRtreeO,paramD,originFamiliesO,paramD['rootFocalClade'],islandFormationSummaryF)
        
def printAnalysisWrapper(paramD,speciesTreeFN,rootFocalClade):
    """Wrapper to run analysis. We take speciesTreeFN and rootFocalClade as
arguments so we can pass in different things in different contexts
(e.g. in xlMode).
    """
    ## setup output directory and file names
    # if directory for analysis doesn't exist yet, make it
    analDir = paramD['analysisDir']
    if glob.glob(analDir)==[]:
        os.mkdir(analDir)

    islandFNStem = paramD['islandsFNStem']
    islandsSummaryFN = os.path.join(analDir,islandFNStem+"Summary.txt")
    islandsTsvFN = os.path.join(analDir,islandFNStem+".tsv")
    genesFNstem = os.path.join(analDir,paramD['genesFNstem'])
    
    ## load stuff
    speciesRtreeO,subtreeD = loadTreeRelatedData(speciesTreeFN)
    strainNamesT=speciesRtreeO.leaves()
    geneOrderD=genomes.createGeneOrderD(paramD['geneOrderFN'],strainNamesT)
    genesO = genomes.genes(paramD['geneInfoFN'])
    genesO.initializeGeneInfoD(paramD['geneInfoFN'],strainNamesT)
    scoresO = scores.readScores(strainNamesT,paramD['scoresFN'])
    originFamiliesO = families.readFamilies(paramD['originFamilyFN'],speciesRtreeO,genesO,"origin")
    islandByNodeD=islands.readIslands(paramD['islandOutFN'],speciesRtreeO)
    
    ## analysis

    # Print islands summary file
    with open(islandsSummaryFN,'w') as islandsOutF:
        analysis.vPrintAllLocusIslands(islandByNodeD,speciesRtreeO,rootFocalClade,subtreeD,originFamiliesO,genesO,islandsOutF)

    # print islands tsv file
    with open(islandsTsvFN,'w') as islandsTsvF:
        analysis.printAllLocusIslandsTsv(islandByNodeD,speciesRtreeO,rootFocalClade,originFamiliesO,genesO,islandsTsvF)
    
    # Print species files with all the genes, grouped by contig
    gene2FamIslandD = analysis.createGene2FamIslandD(islandByNodeD,originFamiliesO)
    analysis.printSpeciesContigs(geneOrderD,genesFNstem,".tsv",genesO,gene2FamIslandD,originFamiliesO,rootFocalClade,strainNamesT)

def createIslandBedWrapper(paramD):
    """Wrapper to make output bed files."""

    strainNamesT,genesO,geneOrderD = loadGenomeRelatedData(paramD)
    speciesRtreeO,subtreeD = loadTreeRelatedData(paramD['speciesTreeFN'])
    originFamiliesO = families.readFamilies(paramD['originFamilyFN'],speciesRtreeO,genesO,"origin")
    islandByNodeD = islands.readIslands(paramD['islandOutFN'],speciesRtreeO)
    genesO.initializeGeneInfoD(paramD['geneInfoFN'],strainNamesT)
    genesO.initializeGeneNumToNameD(paramD['geneInfoFN'],strainNamesT)
    
    # get islands organized by strain
    islandByStrainD = islandBed.createIslandByStrainD(strainNamesT,islandByNodeD,originFamiliesO,genesO)

    islandBed.createAllBeds(islandByStrainD,genesO,speciesRtreeO,strainNamesT,paramD,originFamiliesO)

def interactiveAnalysisWrapper(paramD):
    """Enter interactive mode."""

    ## Set up the modules a bit differently for interactive mode
    import code,sys
    from xenoGI.analysis import createGene2FamIslandD,printScoreMatrix,matchFamilyIsland,printLocusIslandNeighb,vPrintLocusIslandsAtNode,printOutsideFamilyScores
    from .xenoGI import families
    
    ## Wrapper analysis functions. For convenience these assume a
    ## bunch of global variables.
    
    def printFam(familyNum,familiesO,fileF=sys.stdout):
        '''This is a wrapper to provide an easy way to print relevant info on
    a family. For ease of use, we take only two arguments, assuming
    all the other required stuff is available at the top
    level. familyNum is the numerical identifier of a family, and
    familiesO is a family object.
        '''

        print("Family",familyNum,file=fileF)
        fam = familiesO.getFamily(familyNum)
        # print out the locus families
        for lfO in fam.getLocusFamilies():
            print("    LocusFamily",lfO.getStr(genesO," "),file=fileF)

        if hasattr(fam,"sourceFam"):
            print(file=fileF)
            print("    Source family",fam.sourceFam,file=fileF)
            print(file=fileF)
            
        # print("Family error score (count of possibly misassigned genes):",familiesO[familyNum].possibleErrorCt,file=fileF)

        print(file=fileF)
        print("Matrix of raw similarity scores [0,1] between genes in the family",file=fileF)
        printScoreMatrix(familyNum,familiesO,genesO,scoresO,'rawSc',fileF)
        print(file=fileF)
        print(file=fileF)

        print("Matrix of core synteny scores [0,1] between genes in the family",file=fileF)
        printScoreMatrix(familyNum,familiesO,genesO,scoresO,'coreSynSc',fileF)
        print(file=fileF)
        print(file=fileF)

        print("Matrix of synteny scores [0,1] between genes in the family",file=fileF)
        printScoreMatrix(familyNum,familiesO,genesO,scoresO,'synSc',fileF)
        print(file=fileF)
        print(file=fileF)

        printOutsideFamilyScores(familyNum,familiesO,genesO,scoresO,fileF)
        print(file=fileF)
        print(file=fileF)

        if hasattr(fam,"printReconByGeneTree"):
            print("Gene tree",file=fileF)
            print(fam.geneTreeO.toNewickStr(),file=fileF)
            print()
            print("Gene tree annotated with reconciliation [branch events | node events]",file=fileF)
            print(fam.getNewickGeneTreeWithReconLabels(genesO),file=fileF)
            print(file=fileF)
            print("Reconciliation of gene tree onto species tree",file=fileF)       
            fam.printReconByGeneTree(genesO,fileF)
        
    def findGene(searchStr,fileF=sys.stdout):
        '''Find information about a gene. Searches all the fields present in
    the geneInfo file, so the search string can be a locus tag, protein ID, a
    common name, or something present in the description. For each
    hit, prints the gene, LocusIsland, initialFamily, originFamily,
    LocusFamily and gene description. This is a wrapper that assumes
    various required objects are present at the top level.
        '''
        L=matchFamilyIsland(genesO,gene2FamIslandD,searchStr)
        for geneName,locusIslandNum,ifamNum,ofamNum,locusFamNum,descrip in L:
            print("<gene:"+str(geneName),"locIsl:"+str(locusIslandNum),"ifam:"+str(ifamNum),"ofam:"+str(ofamNum),"locFam:"+str(locusFamNum),descrip+">",file=fileF)

    def printLocusIsland(locusIslandNum,synWSize,fileF=sys.stdout):
        '''Print a LocusIsland and its genomic context in each species. We
        include synWSize/2 genes in either direction beyond the locus island.
        '''
        printLocusIslandNeighb(locusIslandNum,synWSize,subtreeD,islandByNodeD,originFamiliesO,geneOrderD,gene2FamIslandD,genesO,paramD['rootFocalClade'],fileF)


    def printLocusIslandsAtNode(node,fileF=sys.stdout):
        '''This is a wrapper to provide an easy way to print all the LocusIslands
    at a particular node in the speciesRtreeO. For ease of use, we take only a node
    number as argument, assuming all the other required stuff is available
    at the top level.
        '''
        vPrintLocusIslandsAtNode(islandByNodeD[node],paramD['rootFocalClade'],subtreeD,originFamiliesO,genesO,fileF)

    ## Load data
    
    strainNamesT,genesO,geneOrderD = loadGenomeRelatedData(paramD)
    speciesRtreeO,subtreeD = loadTreeRelatedData(paramD['speciesTreeFN'])
    genesO.initializeGeneInfoD(paramD['geneInfoFN'],strainNamesT)
    genesO.initializeGeneNumToNameD(paramD['geneInfoFN'],strainNamesT)
    initialFamiliesO = families.readFamilies(paramD['initFamilyFN'],speciesRtreeO,genesO,"initial")
    originFamiliesO = families.readFamilies(paramD['originFamilyFN'],speciesRtreeO,genesO,"origin")
    islandByNodeD=islands.readIslands(paramD['islandOutFN'],speciesRtreeO)
    gene2FamIslandD = createGene2FamIslandD(islandByNodeD,originFamiliesO)
    scoresO = scores.readScores(strainNamesT,paramD['scoresFN'])
    scoresO.createNodeConnectD() # make nodeConnectD attribute
    
    # set up interactive console
    vars = globals()
    vars.update(locals())
    readline.set_completer(rlcompleter.Completer(vars).complete)
    readline.parse_and_bind("tab: complete")
    code.InteractiveConsole(vars).interact()

def debugWrapper(paramD):
    '''Take us into interactive mode for debugging.'''

    ## Set up the modules a bit differently for interactive mode
    import code,sys,numpy
    from .xenoGI import analysis,parameters,trees,genomes,families,islands,analysis,Score,scores

    sys.path.insert(0,os.path.join(sys.path[0],'/biology/bush/htrans/xenoGI.work/misc/'))
    import xlMode
    
    strainNamesT,genesO,geneOrderD = loadGenomeRelatedData(paramD)
    speciesRtreeO,subtreeD = loadTreeRelatedData(paramD['speciesTreeFN'])
   
    # set up interactive console
    vars = globals()
    vars.update(locals())
    readline.set_completer(rlcompleter.Completer(vars).complete)
    readline.parse_and_bind("tab: complete")
    code.InteractiveConsole(vars).interact()



import sys,statistics
from .Family import *
from .Island import *
from . import trees
from . import scores
from . import islands

#### Analysis functions

## general

def printTable(L,indent=0,fileF=sys.stdout):
    '''Given tabular data in a list of lists (where sublists are rows)
print nicely so columns line up. Indent is an optional number of blank spaces to put in front of each row.'''
    # get max width for each column
    colMax=[]
    for col in range(len(L[0])):
        mx=0
        for row in L:
            if len(row[col]) > mx:
                mx = len(row[col])
        colMax.append(mx)
    
    # print
    for row in L:
        for col in range(len(row)):
            row[col]=row[col]+' ' * (colMax[col]-len(row[col]))
            
    for row in L:
        printStr = " "*indent + " | ".join(row)
        print(printStr.rstrip(),file=fileF)

def matchFamilyIsland(genesO,gene2FamIslandD,searchStr):
    '''Return the island number, family number, and gene name(s)
associated with searchStr in genesO.geneInfoD. Searches for a match in all
fields of geneInfoD.'''
    # find matching gene names
    geneMatchL=[]
    for geneNum in gene2FamIslandD:
        # use keys of gene2FamIslandD rather than genesO.geneInfoD
        # because they are from scaffold only.
        valueT=genesO.numToGeneInfo(geneNum)
        for value in valueT:
            if type(value)==str:
                if searchStr in value:
                    descrip = valueT[4]
                    geneMatchL.append((geneNum,descrip))
                    break

    # get family numbers and island numbers
    outL=[]
    for geneNum,descrip in geneMatchL:
        geneName = genesO.numToName(geneNum)
        (locusIslandNum,ifamNum,ofamNum,locusFamNum) = gene2FamIslandD[geneNum]
        outL.append((geneName,locusIslandNum,ifamNum,ofamNum,locusFamNum,descrip))
    return outL
        
## Print scores associated with a family

def printScoreMatrix(familyNum,familiesO,genesO,scoresO,scoreType,fileF):
    '''Print a matrix of scores between all the genes in a family given by
familyNum. Scores are provided by scoresO, and we're extracting the
values associated with scoreType in the edges of this graph.
    '''

    familyGeneNumsL = []
    for lfO in familiesO.getFamily(familyNum).getLocusFamilies():
        familyGeneNumsL.extend(lfO.iterGenes())
    
    rowsL = []
    geneNamesL = [genesO.numToName(gn) for gn in familyGeneNumsL]
    rowsL.append([''] + geneNamesL)
    
    for rowi,gn1 in enumerate(familyGeneNumsL):
        row = [genesO.numToName(familyGeneNumsL[rowi])]
        for gn2 in familyGeneNumsL:
            if scoresO.isEdgePresentByEndNodes(gn1,gn2):
                row.append(format(scoresO.getScoreByEndNodes(gn1,gn2,scoreType),".3f"))
            else:
                row.append('-')
        rowsL.append(row)

    printTable(rowsL,indent=2,fileF=fileF)

def printOutsideFamilyScores(familyNum,familiesO,genesO,scoresO,fileF):
    '''Given a family, print scores for all non-family members with a
connection to genes in family. Scores are provided in the network
scoresO.
    '''

    family = familiesO.getFamily(familyNum)
    outsideGeneNumsS = family.getOutsideConnections(scoresO)
    
    rowL = []
    for familyGeneNum in family.iterGenes():
        familyGeneName = genesO.numToName(familyGeneNum)
        for outsideGeneNum in outsideGeneNumsS:
            if scoresO.isEdgePresentByEndNodes(familyGeneNum,outsideGeneNum):
                outsideGeneName = genesO.numToName(outsideGeneNum)
                rawSc=scoresO.getScoreByEndNodes(familyGeneNum,outsideGeneNum,'rawSc')
                synSc=scoresO.getScoreByEndNodes(familyGeneNum,outsideGeneNum,'synSc')
                coreSynSc=scoresO.getScoreByEndNodes(familyGeneNum,outsideGeneNum,'coreSynSc')
                rowL.append([familyGeneName,outsideGeneName,format(rawSc,".3f"),format(synSc,".3f"),format(coreSynSc,".3f")])

    rowL.sort(key=lambda x: x[2],reverse=True) # sort by score
    rowL.insert(0,['----------','-----------','---','---','-------'])
    rowL.insert(0,['Inside fam','Outside fam','Raw','Syn','CoreSyn'])
                
    print("Printing all scores with non-family members",file=fileF)
    printTable(rowL,indent=2,fileF=fileF)

def getEventCountD(originFamiliesO,eventType):
    '''Produce a dict keyed by species node with value a list of counts of
eventType for every locus family at that node.

    '''
    eventCountD = {node:[] for node in originFamiliesO.speciesRtreeO.preorder()}
    for lfO in originFamiliesO.iterLocusFamilies():
        ct = lfO.countEvents(originFamiliesO,eventType)
        if ct != None:
            eventCountD[lfO.lfMrca].append(ct)
    return eventCountD

def getDtlorScoreSummaryD(originFamiliesO,paramD):
    '''Produce a dict keyed by species node with value a list of dtlor
scores for every locus family at that node.

    '''
    dtlorSummaryD = {node:[] for node in originFamiliesO.speciesRtreeO.preorder()}
    for lfO in originFamiliesO.iterLocusFamilies():
        dtlorScore = lfO.dtlorScore(originFamiliesO,paramD)
        if dtlorScore != None and lfO.aabrhHardCore != []:
            dtlorSummaryD[lfO.lfMrca].append(dtlorScore)
    return dtlorSummaryD 
        
def printSummaryD(speciesRtreeO,summaryD):
    '''Print summary stats of values of entries in summaryD (which is
keyed by species node).'''
    for node in speciesRtreeO.preorder():
        L = summaryD[node]
        if L == []:
            print(node,None,None)
        elif len(L) == 1:
            print(node,L[0],None)
        else:
            print(node,round(statistics.mean(L),3),round(statistics.stdev(L),3))
    
## Print all islands at node

def printIslandLSummary(islandL,fileF):
    '''Given a list of locus islands in islandL (ie a list from a single node),
print a simple tabular summary indicating how many locus families they
have.
    '''
    lenL = []
    for isl in islandL:
        lenL.append(len(isl)) # len of island is num families

    # count how many times each length occurs
    lnCtD = {}
    for ln in lenL:
        if ln in lnCtD:
            lnCtD[ln] += 1
        else:
            lnCtD[ln] = 1

    # print out
    printL = []
    row = ['Num LocusFamilies in LocusIsland','Number of occurrences']
    printL.append(row)
    
    for ln,occurrences in sorted(lnCtD.items()):
        printL.append([str(ln), str(occurrences)])

    printTable(printL,indent=8,fileF=fileF)

def vPrintLocusIsland(island,rootFocalClade,subtreeD,familiesO,genesO,fileF):
    '''Verbose print of a locus island.'''

    print("  LocusIsland",island.id,file=fileF)

    # get species nodes subtended by this mrca
    speciesNodesT = subtreeD[island.mrca].leaves()
    
    # put everything in lists.
    printL=[]
    printL.append(['LocusFamily','lfOrig','Family'])
    for node in speciesNodesT:
        printL[0].append(node)

    # print table of family (row) by gene (col)
    for locusFamO in island.iterLocusFamilies(familiesO):
        newRow=[]
        newRow.append(str(locusFamO.locusFamNum))
        newRow.append(locusFamO.origin(familiesO,rootFocalClade))
        newRow.append(str(locusFamO.famNum))
        for node in speciesNodesT:
            entryL = []
            for geneNum in locusFamO.iterGenesByStrain(node):
                infoT=genesO.numToGeneInfo(geneNum)
                geneName,commonName,locusTag,proteinId,descrip,chrom,start,end,strand=infoT
                commonGeneName = "("+commonName+")" if commonName != '' else ''
                entryL.append(geneName + commonGeneName)
                
            if entryL == []:
                entry = ''
            else:
                entry = ",".join(entryL)

            newRow.append(entry)
        printL.append(newRow)
       
    printTable(printL,indent=4,fileF=fileF)

    # print genes and descriptions
    print(file=fileF)
    for locusFamO in island.iterLocusFamilies(familiesO):
        print("    Locus Family",locusFamO.locusFamNum,file=fileF)
        printL = []
        for geneNum in locusFamO.iterGenes():
            infoT=genesO.numToGeneInfo(geneNum)
            geneName,commonName,locusTag,proteinId,descrip,chrom,start,end,strand=infoT
            printL.append([geneName,descrip])

        printTable(printL,indent=4,fileF=fileF)
        print(file=fileF)

    return
    
def vPrintLocusIslandsAtNode(islandL,rootFocalClade,subtreeD,familiesO,genesO,fileF):
    '''Print a list of islands at a single node.'''
    print("  Summary",file=fileF)
    printIslandLSummary(islandL,fileF)
    print("  -------",file=fileF)
    for island in islandL:
        vPrintLocusIsland(island,rootFocalClade,subtreeD,familiesO,genesO,fileF)
        print('  ------',file=fileF)

def vPrintAllLocusIslands(islandByNodeD,speciesRtreeO,rootFocalClade,subtreeD,familiesO,genesO,fileF):
    '''Loop over all nodes in speciesRtreeO, printing islands at each. '''
    focalRtreeO = speciesRtreeO.subtree(rootFocalClade)
    for node in focalRtreeO.preorder():
        print('########################### ',"Locus Islands at node",node,file=fileF)
        print('',file=fileF)
        vPrintLocusIslandsAtNode(islandByNodeD[node],rootFocalClade,subtreeD,familiesO,genesO,fileF)

def printAllLocusIslandsTsv(islandByNodeD,speciesRtreeO,rootFocalClade,familiesO,genesO,fileF):
    '''Loop over all nodes in speciesRtreeO, printing locus islands at each in tsv form. '''

    # header
    headerTxt = """# Format: locusIslandNum <tab> mrca <tab> locusFamilyOriginStr <tab> [Locus Families and genes in locusIsland]
# The remaining part of each line is organized by locus family
# locusFamilyNum,geneA,geneB <tab> locusFamilyNum,geneC,geneD etc.
# for however many locusFamilies there are."""
    print(headerTxt,file=fileF)
    # print locusIslands
    focalRtreeO = speciesRtreeO.subtree(rootFocalClade)
    for node in focalRtreeO.preorder():
        for island in islandByNodeD[node]:
            oStr = island.getLocusFamilyOriginStr(familiesO,rootFocalClade)
            printL=[str(island.id),str(island.mrca),oStr]
            for locusFamO in island.iterLocusFamilies(familiesO):
                lfamGeneL=[str(locusFamO.locusFamNum)]
                for geneNum in locusFamO.iterGenes():
                    infoT=genesO.numToGeneInfo(geneNum)
                    geneName,commonName,locusTag,proteinId,descrip,chrom,start,end,strand=infoT
                    lfamGeneL.append(geneName)
                printL.append(",".join(lfamGeneL))
            print("\t".join(printL),file=fileF)
    return
    
def createGene2FamIslandD(islandByNodeD,originFamiliesO):
    '''Creates a dictionary keyed by gene number which has the
LocusIsland, Family and LocusFamily for each gene.'''
    D = {}
    for islandsAtNodeL in islandByNodeD.values():
        for locusIslandO in islandsAtNodeL:
            for locusFamO in locusIslandO.iterLocusFamilies(originFamiliesO):
                for gene in locusFamO.iterGenes():
                    ofamNum = locusFamO.famNum
                    ifamNum = originFamiliesO.getFamily(ofamNum).sourceFam
                    D[gene] = (locusIslandO.id,ifamNum,ofamNum,locusFamO.locusFamNum)

    return D

## Print species files with all the genes, grouped by contig

def printSpeciesContigs(geneOrderD,fileStemStr,fileExtensionStr,genesO,gene2FamIslandD,originFamiliesO,rootFocalClade,strainNamesT):
    '''This function produces a set of species specific genome
files. These contain all the genes in a strain laid out in the order
they occur on the contigs. Each gene entry includes LocusIsland and
LocusFamily information, as well as a brief description of the gene's
function.
    '''
    for strainName in strainNamesT:
        if strainName in geneOrderD:
            contigT = geneOrderD[strainName]
            with open(fileStemStr+'-'+strainName+fileExtensionStr,'w') as fileF:
                headerStr = """# Lines not beginning with # have the following format: Gene name <tab> family origin <tab> gene history string <tab> locusIsland <tab> initialFamily <tab> originFamily <tab> locusFamily <tab> locFamMRCA <tab> gene description."""
                print(headerStr,file=fileF)
                for contig in contigT:
                    print("########### Begin contig",file=fileF)
                    printGenes(contig,genesO,gene2FamIslandD,originFamiliesO,rootFocalClade,fileF=fileF)
    return
    
## Print neighborhood of an island

def printLocusIslandNeighb(islandNum,synWSize,subtreeD,islandByNodeD,originFamiliesO,geneOrderD,gene2FamIslandD,genesO,rootFocalClade,fileF):
    '''Print the neighborhood of an island. We include the genes in the island and synWSize/2 genes in either direction.'''

    def printCoordinates(labelText,genesO,firstGene,lastGene,fileF):
        chrom = genesO.numToGeneInfo(firstGene)[5]
        startPos = genesO.numToGeneInfo(firstGene)[6]
        endPos = genesO.numToGeneInfo(lastGene)[7]
        print("    "+labelText,chrom+":"+str(startPos)+"-"+str(endPos),file=fileF)


    print("  LocusIsland:",islandNum,file=fileF)
    
    genesInEitherDirec = int(synWSize/2)

    # get the island object for this islandNum
    for listOfIslands in islandByNodeD.values():
        _,island = islands.searchLocIslandsByID(listOfIslands,islandNum)
        if island != None: break

    if island == None:
        raise ValueError("LocusIsland "+str(islandNum)+" not found.")
        
    mrca = island.mrca
    print("  mrca:",mrca,file=fileF)

    for strainName in subtreeD[mrca].leaves():

        print("  In",strainName,file=fileF)

        islandGenesInStrainL = getIslandGenesInStrain(island,strainName,originFamiliesO)

        if islandGenesInStrainL == []:
            print("the island is not found.",file=fileF)
        else:

            neighbGenesL,firstIslandGene,lastIslandGene=getNeighborhoodGenes(strainName,geneOrderD,islandGenesInStrainL,genesInEitherDirec)

            printCoordinates("Coordinates of locus island",genesO,firstIslandGene,lastIslandGene,fileF)
            printCoordinates("Coordinates of region shown",genesO,neighbGenesL[0],neighbGenesL[-1],fileF)

            printGenes(neighbGenesL,genesO,gene2FamIslandD,originFamiliesO,rootFocalClade,islandGenesInStrainL,fileF)

            
            
def getIslandGenesInStrain(island,strainName,familiesO):
    '''Given an island, a strain number, and our families object, return
all the genes in the island for that strain.'''
    genesL=[]
    for locFam in island.iterLocusFamilies(familiesO):
        genesL.extend(locFam.iterGenesByStrain(strainName))
    return genesL

def getNeighborhoodGenes(strainName,geneOrderD,islandGenesInStrainL,genesInEitherDirec):
    '''Find and return genes in the neighborhood of those given in
islandGenesInStrainL. firstGene and lastGene are the first and last
genes in the island itself.'''
    neighbGenesL=[]
    for contig in geneOrderD[strainName]:
        try:
            # get index of all of these. We're assuming they're on
            # the same contig and in the same area.
            indL=[contig.index(gene) for gene in islandGenesInStrainL]
            maxInd = max(indL)
            minInd = min(indL)
            end = maxInd + genesInEitherDirec +1
            st = minInd-genesInEitherDirec if minInd-genesInEitherDirec>0 else 0 # st can't be less than 0
            neighbGenesL=contig[st:end]

            # get gene numbers of first and last genes in island
            firstGene = contig[minInd]
            lastGene = contig[maxInd]
            
            return neighbGenesL,firstGene,lastGene
        except ValueError:
            continue

def printGenesInteractive(neighbGenesL,genesO,gene2FamIslandD,islandGenesInStrainL,familiesO,fileF):
    '''Given a list of contiguous genes, print them out nicely with
information on locus family and locus island etc.'''

    # now print the neighbors
    rowsL=[]
    for geneNum in neighbGenesL:

        infoT=genesO.numToGeneInfo(geneNum)
        geneName,commonName,locusTag,proteinId,descrip,chrom,start,end,strand=infoT

        
        locIslNum,famNum,locFamNum = gene2FamIslandD[geneNum]
        lfMrca = familiesO.getLocusFamily(locFamNum).lfMrca

        # mark genes in the island with a *
        if geneNum in islandGenesInStrainL:
            geneName = '* '+geneName
        else:
            geneName = '  '+geneName

        infoL = [geneName,"locIsl:"+str(locIslNum),"ofam:"+str(famNum),"locFam:"+str(locFamNum),"lfMrca:"+lfMrca,descrip]

        rowsL.append(infoL)

    printTable(rowsL,indent=4,fileF=fileF)

def printGenes(neighbGenesL,genesO,gene2FamIslandD,originFamiliesO,rootFocalClade,islandGenesInStrainL=[],fileF=sys.stdout):
    '''Given a list of contiguous genes, print them out including

       Gene name,geneOrigin,geneHisStr,locusIsland,ifamNum,ofamNum,locFamNum,locFamMRCA,geneDescription.

       If islandGenesInStrainL is passed in, we assume we're in
       interactive mode, and print in a tabular way. In this case we
       also put * next to any genes in that are in
       islandGenesInStrainL. If this list is empty, we we're in print
       analysis, and print tab delimited.
    '''
    # get strains in focal clade
    focalSubRtreeO = originFamiliesO.speciesRtreeO.subtree(rootFocalClade)
    focalCladeStrainsS = set(focalSubRtreeO.leaves())
    rowsL=[]

    if islandGenesInStrainL != []:
        rowsL.append(['geneName','orig','geneHist','locIsl','ifam','ofam','locFam','lfMrca','descrip'])
        
    for geneNum in neighbGenesL:
        strainName = genesO.numToStrainName(geneNum)
        infoT=genesO.numToGeneInfo(geneNum)
        geneName,commonName,locusTag,proteinId,descrip,chrom,start,end,strand=infoT

        # mark off genes in island (if interactive mode)
        if geneNum in islandGenesInStrainL:
            geneName = '* '+geneName
        else:
            geneName = '  '+geneName

        # family info
        locIslNum,ifamNum,ofamNum,locFamNum = gene2FamIslandD[geneNum]
        famO = originFamiliesO.getFamily(ofamNum)
        
        # only print geneHistory,orign if gene is in focal clade
        if strainName in focalCladeStrainsS:
            geneHisStr = famO.getGeneHistoryStr(str(geneNum)) # geneHistoryD has string keys
            geneOrigin = famO.origin(originFamiliesO.speciesRtreeO,rootFocalClade)
        else:
            geneHisStr = ""
            geneOrigin = ""
            
        lfMrca = originFamiliesO.getLocusFamily(locFamNum).lfMrca

        infoL = [geneName,geneOrigin,geneHisStr,str(locIslNum),str(ifamNum),str(ofamNum),str(locFamNum),lfMrca,descrip]

        rowsL.append(infoL)
        

    # now print
    if islandGenesInStrainL != []:
        # interactive
        printTable(rowsL,indent=4,fileF=fileF)
    else:
        for infoL in rowsL:
            print("\t".join(infoL),file=fileF)
        
    return



# Functions for loading genes and gene order
import sys,glob
from . import fasta
from . import trees

def loadSeq(paramD,fileEnding,genesS=None):
    '''Given paramD and the type of sequence, load the sequences and store
in a dictionary keyed by gene number. fileEnding is a string, either
"_prot.fa" or "_dna.fa". genesS is a set specifying a subset of genes
we want to keep. If missing, we keep all.
    '''
    fileEndingFound = False
    seqD={}
    for fn in glob.glob(paramD['fastaFilePath']):
        if fileEnding in fn:
            fileEndingFound = True # we've seen at least once
            for header,seq in fasta.load(fn):
                gn = int(header.split("_")[0][1:])
                if genesS == None:
                    seqD[gn]=seq
                elif gn in genesS:
                    seqD[gn]=seq

    if not fileEndingFound:
        # we never saw the file ending, the dict is empty.
        raise OSError("There are no fasta files ending in "+fileEnding)    
    return seqD

class genes:
    def __init__(self, geneInfoFN):
        '''genes object. Keeps track of genes present (organized by strain). Also can optionally load a dictionary which allows us to interconvert from numerical to strain represenatations of a gene.'''

        '''
  For later.
 Contains only a single dictionary to convert
one way, either from geneName to number or visa versa. The method
flipDict inverts keys and values in the dictionary.
        '''

        self.geneRangeByStrainD = {}
        self.numGenes = 0
        self.strainGeneRangeT = ()
        self.geneNumToNameD = None
        self.geneInfoD = None
        
        self.initializeGeneRangeByStrainD(geneInfoFN)
        self.initializeStrainGeneRangeT()

    def initializeGeneRangeByStrainD(self,geneInfoFN):
        '''Given a geneInfo file, fill a dict keyed by strain name with values
giving the range of gene numbers in the strain.'''

        f = open(geneInfoFN,'r')

        # handle first line, which gives first strain
        strainName = f.readline().rstrip()[2:]
        enteringNewSpecies = True
        
        while True:
            s = f.readline()
            if s == '':
                break
            elif s[0] == '#':

                # save previous block
                strainRangeEnd = geneNum+1 # make it the number after the last we saw
                self.geneRangeByStrainD[strainName] = (strainRangeStart,strainRangeEnd)

                # get set for next block
                strainName = s.rstrip()[2:] # strainName for next block
                enteringNewSpecies = True
                
            else:
                geneNum,geneName,commonName,locusTag,proteinId,descrip,chrom,start,end,strand=s.rstrip().split('\t')
                geneNum = int(geneNum)
                if enteringNewSpecies == True:
                    strainRangeStart = geneNum
                enteringNewSpecies = False

        f.close()

        # get last strain
        strainRangeEnd = geneNum+1
        self.geneRangeByStrainD[strainName] = (strainRangeStart,strainRangeEnd)
        self.numGenes = strainRangeEnd

    def initializeStrainGeneRangeT(self):
        '''The numToStrainName function works best with a tuple version of the
strain ranges. This method intializes that.'''
        strainRangeL = [(strain,end) for strain,(start,end) in self.geneRangeByStrainD.items()]
        strainRangeL.sort(key=lambda x: x[1]) # sort by end
        self.strainGeneRangeT = tuple(strainRangeL)

    def initializeGeneNumToNameD(self,geneInfoFN,strainNamesL=None):
        '''Given geneInfo file, fill a gene dict keyed by geneNum with value
geneName. If strainNamesL is provided, only load from those strains
present in it. This is an optional part of this data structure, which
isn't loaded by default.
        '''
        self.geneNumToNameD = self.loadDictFromGeneInfoFile(geneInfoFN,strainNamesL,True)

    def initializeGeneInfoD(self,geneInfoFN,strainNamesL=None):
        '''Read gene info from file, returning a dict keyed by gene name with
    information such as description, start position and so on.
        '''
        self.geneInfoD = self.loadDictFromGeneInfoFile(geneInfoFN,strainNamesL,False)

    def loadDictFromGeneInfoFile(self,geneInfoFN,strainNamesL,onlyNames):
        '''Reads in a dictionary from geneInfo file. Keys are geneNum. If
onlyNames is True, then we make the values geneName, otherwise all
other fields.'''

        D = {}
        f = open(geneInfoFN,'r')
        while True:
            s = f.readline()
            if s == '':
                break
            elif s[0] == '#':
                strainName = s.rstrip()[2:]
                if strainNamesL == None or strainName in strainNamesL:
                    keepGenes = True
                else:
                    keepGenes = False
            else:
                geneNum,geneName,commonName,locusTag,proteinId,descrip,chrom,start,end,strand=s.rstrip().split('\t')
                if keepGenes == True:
                    geneNum = int(geneNum)
                    if onlyNames:
                        D[geneNum] = geneName
                    else:
                        D[geneNum] = geneName,commonName,locusTag,proteinId,descrip,chrom,start,end,strand
                        
        f.close()
        return D

    def numToStrainName(self,geneNum):
        '''Given a gene number, return the strain name corresponding. Uses
binary search to find the where geneNum falls in strainGeneRangeT.'''

        st = 0
        end = len(self.strainGeneRangeT)
        
        while True:
            mid = (st + end) // 2
            if self.strainGeneRangeT[mid][1] <= geneNum:
                st = mid + 1
            else:
                # self.strainGeneRangeT[mid][1] > geneNum, so two
                # possibilities, we're in the range, or we're too high
                if self.strainGeneRangeT[mid-1][1] < geneNum:
                    # we're in the range
                    return self.strainGeneRangeT[mid][0]
                else:
                    end = mid
                    if end<=st:
                        return self.strainGeneRangeT[end][0]

    def numToName(self,geneNumber):
        '''Given gene number, return gene name. Assumes self.geneNumToNameD
has been initialized.'''
        return self.geneNumToNameD[geneNumber]

    def iterGenes(self,strainL=None):
        '''Iterate over genes. If strainL is None (or is not given) iterates
over all genes. If strainL is given, iterates over genes from only the
strains present in strainL.'''

        for strainName,endRange in self.strainGeneRangeT:
            # this will give in order of gene number
            if strainL == None or strainName in strainL:
                for geneNum in self.iterGenesStrain(strainName):
                    yield geneNum

    def iterGenesStrain(self,strainName):
        '''Iterates over all genes in strainName.'''
        rangeStart,rangeEnd = self.geneRangeByStrainD[strainName]
        return range(rangeStart,rangeEnd)

    def numToGeneInfo(self,geneNumber):
        '''Given gene number, return other info about gene. Assumes
self.geneInfoD has been initialized.
        '''
        return self.geneInfoD[geneNumber]
    
    def __len__(self):
        return self.numGenes
    
    def __repr__(self):
        return "<genes object with "+str(len(self))+" genes.>"
        
def getProximityInWindow(geneWinT,geneProximityD):
    '''Given a window of genes, calculate the distances between first gene
and the rest (in number of genes) and store in geneProximityD. Updates
geneProximityD in place, returning None.
    '''
    gnA=geneWinT[0]
    for i in range(1,len(geneWinT)):
        gnB=geneWinT[i]

        if gnA<gnB: # always put lower gene number first
            geneProximityD[(gnA,gnB)]=i
        else:
            geneProximityD[(gnB,gnA)]=i

def createGeneProximityD(geneOrderD,geneProximityForGroup):
    '''Go though a gene order tuple pulling out pairs of genes that are
within geneProximityForGroup genes of each other. Store in a dict keyed by
gene pair, with value equal to the distance between the two genes,
measured in number of genes.'''
    geneProximityD = {}
    for contigT in geneOrderD.values():
        for geneNumT in contigT:
            for i in range(len(geneNumT)):
                # slide over genes w/win geneProximityForGroup+1
                # wide.
                getProximityInWindow(geneNumT[i:i+(geneProximityForGroup+1)],geneProximityD)
                    
    return geneProximityD

def createGeneOrderD(geneOrderFN,strainNamesL):
    '''Go though gene order file and get orderings into a set of
tuples. Put in a dict, keyed by strain name. We include only the
strains in strainNamesL. If strainNamesL is None, load all.

    '''
    geneOrderD = {}
    f = open(geneOrderFN,'r')
    while True:
        s = f.readline()
        if s == '':
            break
        s=s.rstrip()
        # note, our gene order format has contigs separated by \t, and
        # genes within them separated by a space character.
        L=s.split('\t')
        strain = L[0]
        if strainNamesL == None or strain in strainNamesL:
            # we only load those strains that are in strainNamesL (or all
            # if strainNamesL is None)
            contigL=[]
            for contig in L[1:]:
                geneNumT=tuple((int(g) for g in contig.split(' ')))
                contigL.append(geneNumT)
            geneOrderD[strain] = tuple(contigL)
    f.close()
    return geneOrderD




from Bio import Phylo
from collections import OrderedDict
from xenoGI import trees

## Globals

ROOT_PARENT_NAME = "" # name for the parent of the root node in rooted trees

## Classes

class Tree:
    def __init__(self, nodeConnectD=None):
        '''Base class to be inherited by Rtree and Utree. Key internal data
structure is nodeConnectD, which is keyed by node and has values
(parent, child1, child2...). All nodes names are represented as
strings. In the case of a gene tree, the tip may be a gene number, but
we still hold it as a string.

        '''
        self.nodeConnectD = nodeConnectD
        self.leafNodeT = None
        self.internalNodeT = None
        self.preOrderT = None
        self.branchLenD = None
        
        if self.nodeConnectD != None:
            self.__updateSecondaryAttributes__()

    def fileStr(self):
        '''Return a string representation of a Tree object. Outer separator is "|".

        rootNode/arbitraryNode | node connection info | branch len info

        The first position gives the root node in a rooted tree, or
        else the arbitraryNode in an unrooted tree.

        Node connection info is organized as follows. Each entry in
        nodeConnectD gets internally separated by spaces. Different
        entries are separated by commas.

        Branch connection info is organized similarly. Each entry in
        branchLenD gets internally separated by spaces. Different
        entries are separated by commas. If branch len info
        is absent, will be None.
        '''
        outL = []

        # root/arbitary node
        if hasattr(self,"rootNode"):
            # rooted tree
            outL.append(str(self.rootNode))
        else:
            # unrooted
            outL.append(str(self.arbitraryNode))

        # node connection
        nodeConnecL=[] # first thing is root node.
        for node,connecT in self.nodeConnectD.items():
            # elements in one entry of nodeConnectD separated by spaces
            entry = node + " " + " ".join(connecT)
            nodeConnecL.append(entry)
        nodeConnecStr = ",".join(nodeConnecL)
        outL.append(nodeConnecStr)

        # branches
        if self.branchLenD == None:
            outL.append("None")
        else:
            branchLenL = []
            for branchPair,brLen in self.branchLenD.items():
                entry = branchPair[0] + " " + branchPair[1] + " " + str(brLen)
                branchLenL.append(entry)
            branchLenStr = ",".join(branchLenL)
            outL.append(branchLenStr)
            
        return "|".join(outL)

    def populateAttributes(self,nodeConnectD,rootArbNode,branchLenD=None):
        '''Populate the attributes of an empty tree object.'''

        if self.nodeConnectD != None or self.branchLenD != None:
            raise ValueError("Attempting to populate attributes of a non-empty tree.")    

        self.nodeConnectD = nodeConnectD
        self.__updateSecondaryAttributes__()
        self.preOrderT = self.__traversePreOrder__(rootArbNode)
        
        # root/arbitraryNode
        if isinstance(self, Rtree):
            self.rootNode = rootArbNode
        else:
            self.arbitraryNode = rootArbNode
            
        # branches
        if branchLenD == None:
            self.branchLenD = branchLenD
        else:
            self.branchPairT = self.__createBranchPairT__()
            self.branchLenD = {}
            for branchPair in self.branchPairT:
                if branchPair in branchLenD:
                    self.branchLenD[branchPair] = branchLenD[branchPair]
                else:
                    # flip order
                    self.branchLenD[branchPair] = branchLenD[(branchPair[1],branchPair[0])]
        
    def fromString(self,treeStr):
        '''Populate attributes by parsing the string treeStr (which has likely
been read from a file). This method allows multifurcating trees. See
fileStr method for description of format.

        '''
        rootArbNode,nodeConnecStr,branchLenStr = treeStr.split("|")

        # nodeConnectD
        L = nodeConnecStr.split(",")
        nodeConnectD = {}
        for entryStr in L:
            entryL = entryStr.split(" ")
            node=entryL[0]
            connecT=tuple(entryL[1:])
            nodeConnectD[node] = connecT

        # branchLenD
        if branchLenStr == "None":
            branchLenD = None
        else:
            L = branchLenStr.split(",")
            branchLenD = {}
            for entryStr in L:
                br0,br1,brLen = entryStr.split(" ")
                branchLenD[(br0,br1)] = float(brLen)

        # put in these attributes
        self.populateAttributes(nodeConnectD,rootArbNode,branchLenD)
                
    def isLeaf(self,node):
        '''Return boolean if node is a leaf'''
        if len(self.nodeConnectD[node]) < 3:
            # internal node must have at least 3 connections
            return True
        else: return False

    def preorder(self):
        return self.preOrderT

    def leaves(self):
        return self.leafNodeT

    def internals(self):
        return self.internalNodeT

    def leafCount(self):
        return len(self.leafNodeT)

    def internalNodeCount(self):
        return len(self.internalNodeT)
    
    def nodeCount(self):
        return len(self.nodeConnectD)

    def multifurcatingNodes(self):
        '''Return a list of all nodes where there's a multifurcation (more
than 3 branches).'''
        multifurcL = []
        for node,connecT in self.nodeConnectD.items():
            if len(connecT) > 3:
                multifurcL.append(node)
        return multifurcL

    def getParent(self,node):
        '''For rooted trees, give parent. For unrooted trees, this simply
gives a node leading to this node. We have this method also for
unrooted trees since it can be useful for finding what leads to a
tip.
        '''
        connecT = self.nodeConnectD[node]
        return connecT[0] # parent is first element

    def binarize(self,gtLocusMapD=None):
        '''Arbitarily convert a tree with multifurcating nodes into one with
binary nodes. Newly introduced nodes have a "b" in their naming to
mark them off. If tree has branch lengths, then newly introduced
branches are given 0 length. gtLocusMapD gives the syntenic regions of
genes. If this is not None, then we use it to help order the way we
binarize multifurcating nodes.
        '''
        
        # funcs
        def getNewNodes(newNodeCtr,parent,node,childL,updateParentD):
            '''Goes through a list of children from a multifurcating node, and
creates a series of clades by having one group branch off at each
step. First groups listed will be first to branch off. Returns a list
of new nodes, and also an updateParentD which can later be used to
make sure the parents are correct.
            '''
            if len(childL) == 2:
                connecT = (parent,childL[0],childL[1])
                return newNodeCtr,updateParentD,[(node,connecT)],
            else:
                newNode = "b"+str(newNodeCtr)
                newNodeCtr+=1
                connecT = (parent,childL[0],newNode)
                # put rest of children in updateParentD
                for child in childL[1:]:
                    updateParentD[child] = newNode # for now they descent from newNode
                # recurse
                newNodeCtr,updateParentD,L=getNewNodes(newNodeCtr,node,newNode,childL[1:],updateParentD)
                return newNodeCtr,updateParentD,[(node,connecT)]+L

        def orderChildrenBySynteny(childT,gtLocusMapD):
            '''Given a tuple of children from a multifurcating node, and a dict of
 syntenic regions for these, reorder the children in a way that makes
 use of synteny. When we construct the binary tree, the first listed
 children in our output will branch off first. If there are branches
 in childT where synteny is the same in all genes, and where this
 syteny is not shared with other child branches, then we'd like to put
 those first. This tweak is intented to make the binarization more
 likely to reflect what really happened.
            '''
            childToSynRegTupleD = {}
            for child in childT:
                # get the synReg for this child
                if self.isLeaf(child):
                    synRegT = (gtLocusMapD[child],)
                else:
                    # internal node
                    synRegS = set()
                    for chLeaf in self.subtree(child).leaves():
                        synRegS.add(gtLocusMapD[chLeaf])

                    synRegT = tuple(sorted(synRegS))
                    
                childToSynRegTupleD[child] = synRegT

            # create dict keyed by synReg
            synRegToChildD = {}
            for child in childToSynRegTupleD:
                for synReg in childToSynRegTupleD[child]:
                    if synReg in synRegToChildD:
                        synRegToChildD[synReg].append(child)
                    else:
                        synRegToChildD[synReg] = [child]

            # for each child, characterize the max num children its
            # synRegs appear in
            outChildL = []
            for child in childT:

                synRegT = childToSynRegTupleD[child]
                maxAppearances = 0
                for synReg in synRegT:
                    if len(synRegToChildD[synReg]) > maxAppearances:
                        maxAppearances = len(synRegToChildD[synReg])
                
                outChildL.append((maxAppearances,child))
                
            # sort by max appearances, so the ones with less (ie 1) come first
            outChildL.sort(key = lambda x: x[0])

            return tuple((child for _,child in outChildL))
            
        # main part of binarize

        # get node connections
        newNodeConnectD = {}
        updateParentD = {}
        newNodeCtr = 0
        for node,connecT in self.nodeConnectD.items():
            if len(connecT) <= 3:
                # not multifurcating
                newNodeConnectD[node] = connecT
            else:
                # multifurcating        
                parent = connecT[0]
                childT = connecT[1:]

                if gtLocusMapD != None:
                    childT = orderChildrenBySynteny(childT,gtLocusMapD)
                
                newNodeCtr,updateParentD,newNodeL = getNewNodes(newNodeCtr,parent,node,childT,updateParentD)
                for nNode,nConnecT in newNodeL:
                    newNodeConnectD[nNode] = nConnecT

        # fix newNodeConnectD using updateParentD
        for node in updateParentD.keys():
            connecT = newNodeConnectD[node]
            newConnecT = (updateParentD[node],) + connecT[1:] # new parent
            newNodeConnectD[node] = newConnecT
            
        # get branch lens
        if self.branchLenD == None:
            newBranchLenD = None
        else:
            newBranchLenD = {}
            for nNode,nConnecT in newNodeConnectD.items():
                for otherEndOfBranch in nConnecT:
                    if (nNode,otherEndOfBranch) in self.branchLenD:
                        newBranchLenD[(nNode,otherEndOfBranch)] = self.branchLenD[(nNode,otherEndOfBranch)]
                    elif (otherEndOfBranch,nNode) in self.branchLenD:
                        # it could be either order
                        newBranchLenD[(otherEndOfBranch,nNode)] = self.branchLenD[(otherEndOfBranch,nNode)]
                    else:
                        newBranchLenD[(nNode,otherEndOfBranch)] = 0
            
                                
        # create output tree
        if hasattr(self,"rootNode"):      
            # we're a rooted tree
            outTreeO = Rtree()
            rootArbNode = self.rootNode
        else:
            outTreeO = Utree()
            rootArbNode = self.arbitraryNode

        outTreeO.populateAttributes(newNodeConnectD,rootArbNode,newBranchLenD)
        return outTreeO
        
    def __contains__(self,node):
        return node in self.nodeConnectD
        
    def __updateSecondaryAttributes__(self):
        '''Create leafNodeT and internalNodeT.'''
        leafNodeT,internalNodeT = self.__updateSecondaryAttributesHelper__(self.nodeConnectD)
        self.leafNodeT = leafNodeT
        self.internalNodeT = internalNodeT

    def __updateSecondaryAttributesHelper__(self,nodeConnectD):
        '''Actually does the work of getting internals and leaves..'''
        leafNodeL = []
        internalNodeL = []
        for node,connecT in nodeConnectD.items():
            if len(connecT) < 3:
                # it's a tip
                leafNodeL.append(node)
            else:
                internalNodeL.append(node)
        return tuple(leafNodeL),tuple(internalNodeL)
        
    def iterBranches(self):
        '''Iterate over all branches, returning branchPair tuple and branch
length.'''
        if self.branchLenD == None:
            raise ValueError("This unrooted tree object does not have branch lengths defined.")
            
        for branchPair in self.branchPairT:
            branchLen = self.branchLenD[branchPair]
            yield branchPair,branchLen

    def getBranchesByLengthL(self):
        '''Returns a list of tuples (branchPair,branchLen) sorted by branch length. Largest first.'''
        if self.branchLenD == None:
            raise ValueError("This unrooted tree object does not have branch lengths defined.")
            
        L=list(self.branchLenD.items())
        L.sort(key=lambda x: x[1],reverse=True)
        return L
            
    def maxBranchLen(self):
        '''Find the maximum branch length present, and return the
corresponding branchPair tuple and the length.

        '''
        L=list(self.branchLenD.items())
        L.sort(key=lambda x: x[1])
        return L[-1]
        
    def __traversePreOrder__(self,node):
        '''Traverse in preorder starting at node'''
        if self.nodeCount() == 2:
            # two tip tree is a special case            
            return self.leafNodeT
        else:
            return tuple(self.__traversePreOrderNodeConnectD__(self.nodeConnectD,node,ROOT_PARENT_NAME))

    def __traversePreOrderNodeConnectD__(self,D,node,parentNode):
        '''Traverse nodeConnectD starting at node. Do not recurse along
parentNode.
        '''
        connecT=D[node]
        if len(connecT)==1:
            return [node]
        else:
            outL = [node]
            for child in connecT:
                if child != parentNode:
                    tempL = self.__traversePreOrderNodeConnectD__(D,child,node)
                    outL = outL + tempL
            return outL

    def __traverseForNewickStr__(self,node,parentNode,includeBrLength,nodeLabelD):
        '''Produce newick representation of self starting at node. If
includeBrLength is True, include branch lengths. nodeLabelD is a
dictionary containing labels to be added to the nodes. If not to be
used, these should be set to None.
        '''
        connecT=self.nodeConnectD[node]

        # construct node label
        if nodeLabelD == None:
            nodeLab = node
        else:
            nodeLab = node+"["+nodeLabelD[node]+"]"
            
        # construct branch label
        if includeBrLength and parentNode != ROOT_PARENT_NAME:
            brLab = ":"+str(self.branchLenD[(parentNode,node)])
        else:
            brLab = ""
        
        # construct newick    
        if len(connecT)==1: # tip
            return nodeLab+brLab
        else:
            outL = []
            for child in connecT:
                if child != parentNode:
                    newickStr = self.__traverseForNewickStr__(child,node,includeBrLength,nodeLabelD)
                    outL.append(newickStr)
            return "("+ ",".join(outL)+")"+nodeLab+brLab
        
    def __createBranchPairT__(self):
        '''Make the branchPairT attribute to represent the branches. This is a
tuple of tuples, where the subtuples represent edges and consist of
the two nodes on either end.
        '''
        S = set()
        for node,connecT in self.nodeConnectD.items():
            for otherNode in connecT:
                # don't include any edge with ROOT_PARENT_NAME
                if otherNode != ROOT_PARENT_NAME:
                    # put them in preorder
                    edgeT = tuple((nd for nd in self.preorder() if nd in [node,otherNode]))
                    S.add(edgeT)
        return tuple(sorted(S))
        
    def __eq__(self,other):
        '''See if two trees are equivalent.'''
        if hasattr(self,"rootNode") and  hasattr(other,"rootNode"):
            # both rooted
            if self.rootNode != other.rootNode: return False
        elif hasattr(self,"arbitraryNode") and  hasattr(other,"arbitraryNode"):
            # both unrooted
            if self.arbitraryNode != other.arbitraryNode: return False
        else:
            # different types of tree
            return False

        # if we made it here, they are same type of tree, and basis
        # node matches

        # check node connections
        if sorted(self.nodeConnectD.items()) != sorted(other.nodeConnectD.items()):
            return False

        # check branch lengths if they exist
        if self.branchLenD == other.branchLenD:
            return True
        else:
            # don't match
            if self.branchLenD == None or other.branchLen == None:
                # one is None
                return False
            else:
                # both are dicts. Check more carefully
                selfL = []
                for brT,brLen in self.branchLenD.items():
                    selfL.append((brT,round(brLen,4)))
                otherL = []
                for brT,brLen in other.branchLenD.items():
                    otherL.append((brT,round(brLen,4)))

                selfL.sort()
                otherL.sort()
                return selfL == otherL
        
class Rtree(Tree):
    def __init__(self, nodeConnectD=None,rootNode=None):
        '''Initialize a rooted tree object.'''
        super().__init__(nodeConnectD)
        self.rootNode = rootNode

        if self.nodeConnectD != None:
            self.preOrderT = self.__traversePreOrder__(self.rootNode)
        
    def fromNewickFileLoadSpeciesTree(self,treeFN,outGroupTaxaL=None,includeBrLen=False):
        '''Populate attributes based on newick file in treeFN. This method
assumes we are working with a species tree. It must be rooted,
bifurcating and have named internal nodes. An optional argument
outGroupTaxaL can be used to provide a list of outgroups to root the
tree. If outGroupTaxaL is present, we assume we are being asked to
prepare a species tree (e.g. from ASTRAL output) and we do other
preparation such as naming internal nodes).

        '''
        if outGroupTaxaL == None:
            bpTree = Phylo.read(treeFN, 'newick', rooted=True)
        else:
            bpTree = Phylo.read(treeFN, 'newick')
            bpTree = trees.prepareTree(bpTree,outGroupTaxaL)
            
        self.__checkSpeciesTree__(bpTree)

        nodeConnectD,branchLenD = self.__bioPhyloToNodeConnectD__(bpTree)

        if includeBrLen:
            self.populateAttributes(nodeConnectD,bpTree.root.name,branchLenD)
        else:
            self.populateAttributes(nodeConnectD,bpTree.root.name)

    def toNewickStr(self,includeBrLength=False,nodeLabelD=None):
        '''Output a newick string.'''
        return self.__traverseForNewickStr__(self.rootNode,ROOT_PARENT_NAME,includeBrLength,nodeLabelD)
        
    def subtree(self,node):
        '''Return a new Rtree object with the subtree rooted at node.'''
        
        def traverse(D,subD,node):
            '''Get nodeConnectD for subtree'''
            connecT=D[node]
            subD[node] = connecT
            if len(connecT)==1:
                return
            else:
                for child in connecT[1:]: # 1st entry is parent
                    traverse(D,subD,child)
                return

        subD = {}
        traverse(self.nodeConnectD,subD,node)

        # put special value in parent of root position
        oldConnecT = subD[node]
        newConnecT = (ROOT_PARENT_NAME,)+oldConnecT[1:]
        subD[node] = newConnecT

        # make tree
        outTreeO = Rtree()

        # if we have branch lens, include these
        if self.branchLenD == None or len(subD)<2:
            # if self has no branch lengths, or if subtree is a tip, then branchLenD will be None
            outTreeO.populateAttributes(subD,node)
        else:
            subBranchLenD = {}
            for tempNode,connecT in subD.items():
                parentNode = connecT[0]
                if parentNode != ROOT_PARENT_NAME:
                    subBranchLenD[(parentNode,tempNode)] = self.branchLenD[(parentNode,tempNode)]

            outTreeO.populateAttributes(subD,node,subBranchLenD)

        return outTreeO
        
    def createSubtreeD(self):
        '''Get all subtrees and put them in a dict keyed by node name.'''
        subtreeD = {}
        for node in self.preorder():
            subtree = self.subtree(node)
            subtreeD[node] = subtree
        return subtreeD
            
    def children(self,node):
        '''Return tuple of children of node.'''
        connecT = self.nodeConnectD[node]
        return connecT[1:]
        
    def ancestors(self,node):
        '''Return a tuple of nodes ancestral to node.'''

        def traverse(D,node):
            connecT=D[node]
            parent = connecT[0]
            if parent == ROOT_PARENT_NAME: # at root
                return []
            else:
                return [parent]+traverse(D,parent)

        return tuple(traverse(self.nodeConnectD,node))

    def getNearestNeighborL(self,leaf):
        '''Given a leaf, return a list containing the other leaf or
    leaves which are most closely related to leaf.'''
        parent = self.getParent(leaf) # assume leaf is in tree
        subRtreeO = self.subtree(parent)
        leafL = list(subRtreeO.leaves())
        leafL.remove(leaf)
        return leafL

    def findMrcaPair(self,node1,node2):
        '''Return the most recent common ancestor of node1 and node2.'''

        anc1L = (node1,) + self.ancestors(node1)
        anc2L = (node2,) + self.ancestors(node2)

        for node in anc1L:
            if node in anc2L:
                return node
        return None

    def findMrca(self,nodeL):
        '''Get mrca of a list of nodes'''
        mrca = nodeL[0]
        for node in nodeL[1:]:
            mrca = self.findMrcaPair(mrca,node)
        return mrca

    def prune(self,leafL):
        '''Prune down to a tree with only the leaves in leafL, combining
branches and removing nodes as needed. Returns a new Rtree.
        '''

        # funcs
        def traverseBlD(subTreeO,activeNodeS,blD,pruneParent,previousBrLen,node):
            '''Traverse subtree, keeping only nodes in activeNodeS and adjusting
            branch lengths (add to blD).
            '''
            if subTreeO.isLeaf(node):
                if pruneParent != ROOT_PARENT_NAME:
                    blD[(pruneParent,node)] = previousBrLen + subTreeO.branchLenD[(subRtreeO.getParent(node),node)]
                return blD
            else:
                activeChildrenL = [child for child in subTreeO.children(node) if child in activeNodeS]
                if len(activeChildrenL) > 1:
                    # we don't need to remove this node
                    if pruneParent != ROOT_PARENT_NAME:
                        # here we really want to pass in subRtreeO.getParent(node)
                        # which is the immediate last one, even if its not included in the pruned tree
                        blD[(pruneParent,node)] = previousBrLen + subTreeO.branchLenD[(subRtreeO.getParent(node),node)]
                    # recurse
                    for child in activeChildrenL:
                        blD = traverseBlD(subTreeO,activeNodeS,blD,node,0,child)
                else:
                    # there is only one active child, and we should remove this node
                    child = activeChildrenL[0]
                    brLenToPassAlong = previousBrLen + subTreeO.branchLenD[(subRtreeO.getParent(node),node)]
                    blD = traverseBlD(subTreeO,activeNodeS,blD,pruneParent,brLenToPassAlong,child)
                    
                return blD
            
        def buildNcD(blD,ncD,parent,node):
            '''Fillout ncD (nodeConnectD) based on the branches and lengths in blD.'''
            # get children of node and put in connecT
            childL = []
            for branchPair in blD:
                if branchPair[0] == node:
                    childL.append(branchPair[1])

            connecT = tuple([parent] + childL)
            ncD[node] = connecT

            # recurse. childL will be empty if node a tip.
            for child in childL:
                ncD = buildNcD(blD,ncD,node,child)

            return ncD
                
        # main section
        mrcaNode = self.findMrca(leafL)
        subRtreeO = self.subtree(mrcaNode)

        activeNodeS = set() # leafL + all ancestors
        for leaf in leafL:
            activeNodeS.add(leaf)
            for node in subRtreeO.ancestors(leaf):
                activeNodeS.add(node)

        blD = {}
        blD = traverseBlD(subRtreeO,activeNodeS,blD,subRtreeO.getParent(mrcaNode),0,mrcaNode)
        ncD = {}
        ncD = buildNcD(blD,ncD,ROOT_PARENT_NAME,mrcaNode)
        
        outRtreeO = Rtree()
        outRtreeO.populateAttributes(ncD,mrcaNode,blD)

        return outRtreeO

    def unroot(self):
        '''Return an Utree object corresponding to self. We assume self has at
least two tips.'''
        # funcs
        def traverse(nodeConnectD,branchLenD,parent,node):
            '''Copy entries into nodeConnecteD, starting with node.'''
            if self.isLeaf(node):
                nodeConnectD[node] = (parent,)
            else:
                childT = self.children(node)
                nodeConnectD[node] = (parent,) + childT
                for child in childT:
                    # get branch to child
                    if branchLenD != None:
                        branchLenD[(node,child)] = self.branchLenD[(node,child)]
                    
                    nodeConnectD,branchLenD = traverse(nodeConnectD,branchLenD,node,child)
                
            return nodeConnectD,branchLenD
                
        # unroot main
        nodeConnectD={}
        if self.branchLenD == None:
            branchLenD = None
        else:
            branchLenD = {}

        childT = self.children(self.rootNode)
        if len(childT) < 2:
            raise ValueError("Trying to unroot a single tip tree.")
        elif len(childT) == 2:
            # handle this first branch
            if branchLenD != None:
                branchLenD[(childT[0],childT[1])] = self.branchLenD[(self.rootNode,childT[0])] + self.branchLenD[(self.rootNode,childT[1])]
            # recurse
            nodeConnectD,branchLenD = traverse(nodeConnectD,branchLenD,childT[0],childT[1])
            nodeConnectD,branchLenD = traverse(nodeConnectD,branchLenD,childT[1],childT[0])
            arbitraryNode = childT[0]
        else:
            # Multifurcation at root. Not clear this will ever come up.
            for child in childT:
                nodeConnectD,branchLenD = traverse(nodeConnectD,branchLenD,node,child)
            arbitraryNode = self.rootNode
            
        utreeO = Utree()
        utreeO.populateAttributes(nodeConnectD,arbitraryNode,branchLenD)
        return utreeO
        
    def createDtlorD(self,spTreeB):
        '''Return a dict in the format expected by the dtlor reconciliation
code. Key is a node, value is (parent, node, child1, child2). For now
this only works for bifurcating trees. spTreeB is a boolean that is
True if this is a species tree, and False if it is for a gene
tree.
        '''
        if self.multifurcatingNodes() != []:
            raise ValueError("This tree is not bifurcating.")
        
        dtlorD = OrderedDict()
        for node in self.preorder():
            connecT = self.nodeConnectD[node]
            if len(connecT) == 1:
                # leaf
                parent = connecT[0]
                dtlorD[node] = (parent, node, None, None)
            else:
                parent,child1,child2 = connecT                
                if node == self.rootNode:
                    if spTreeB:
                        dtlorD[node] = ("h_root", node, child1, child2)
                    else:
                        dtlorD[node] = ("p_root", node, child1, child2)
                else:
                    dtlorD[node] = (parent, node, child1, child2)
                    
        return dtlorD
    
    def __checkSpeciesTree__(self,bpTree):
        '''Check that a biopython tree is rooted, bifurcating, and has named
    internal nodes. Throw error if not. Returns None.
        '''
        # check its bifurcating. This actually ignores the root, but checks
        # for non-bifurcating nodes below that.
        if not bpTree.is_bifurcating():
            raise ValueError("This tree is not bifurcating.")

        # Now check to make sure there are only two clades at the root of
        # this tree. Otherwise, its going to mess up bioPhyloToTupleTree
        if len(bpTree.clade) > 2:
            raise ValueError("readTree requires a rooted tree with exactly two clades at the root. This input tree has more than two. This typically means that it is intended to be read as an unrooted tree.")

        # Make sure the internal nodes are labelled.
        predictedNumIntNodes = bpTree.count_terminals() - 1
        internalL = bpTree.get_nonterminals()
        if len(internalL) != predictedNumIntNodes or any((n.name==None for n in internalL)):
            raise ValueError("All the internal nodes of the input tree must have names.")

        # Make sure all names unique
        nodeL = [n.name for n in bpTree.get_nonterminals() + bpTree.get_terminals()]
        if len(nodeL) != len(set(nodeL)):
            raise ValueError("All nodes of the input tree (tips and interal) must have unique names.")
        
    def __bioPhyloToNodeConnectD__(self,bpTree):
        '''Convert a biopython tree object to a node connection dict with keys
    that are nodes, and values that are (parent, child1, child2...).'''
        nodeConnectD = {}
        branchLenD = {}
        # parent of root is ROOT_PARENT_NAME
        return self.__bioPhyloCladeToNodeConnectD__(bpTree.root,nodeConnectD,branchLenD,ROOT_PARENT_NAME)

    def __bioPhyloCladeToNodeConnectD__(self,clade,nodeConnectD,branchLenD,parent):
        '''Recursive helper to convert a biopython clade object.
        '''
        # get branch, but only if its not branch leading to root
        if parent != ROOT_PARENT_NAME: branchLenD[(parent,clade.name)] = clade.branch_length

        # fill nodeConnectD
        if clade.is_terminal():
            nodeConnectD[clade.name] = (parent,)
            return nodeConnectD,branchLenD
        else:
            nodeConnectD[clade.name] = (parent,clade[0].name,clade[1].name)
            for iterClade in clade:
                nodeConnectD,branchLenD = self.__bioPhyloCladeToNodeConnectD__(iterClade,nodeConnectD,branchLenD,clade.name)
            return nodeConnectD,branchLenD

    def __repr__(self):
        return "Rtree: "+self.toNewickStr()


class Utree(Tree):
    def __init__(self, nodeConnectD=None):
        '''Initialize an unrooted tree object.'''
        super().__init__(nodeConnectD)
        
        if self.nodeConnectD != None:
            # arbitraryNode provides a starting point for some methods
            self.arbitraryNode = self.internals()[0] if len(self.internals())>0 else self.leaves()[0]
            self.preOrderT = self.__traversePreOrder__(self.arbitraryNode)
            self.branchPairT = self.__createBranchPairT__()
            
    def fromNewickFile(self,treeFN):
        '''Populate attributes based on newick file in treeFN. This method
assumes we are working with an unrooted gene tree that does not have
named interal nodes (we will create those).

        '''
        bpTree = Phylo.read(treeFN, 'newick', rooted=False)
        # handle special case of one tip tree
        if bpTree.count_terminals()==1:
            # special case, one tip tree
            nodeConnectD={bpTree.get_terminals()[0].name:()}
            branchLenL=[('','',None)]
        else:
            iNodeNum,nodeConnectD,branchLenL = self.__bioPhyloToNodeConnectD__(bpTree.clade,ROOT_PARENT_NAME,{},0)

        # get arbitrary node
        tempLeavesT,tempInternalsT = self.__updateSecondaryAttributesHelper__(nodeConnectD)
        arbitraryNode = tempInternalsT[0] if len(tempInternalsT)>0 else tempLeavesT[0]

        # branchLenD
        branchLenD = {}
        for parent,child,brLen in branchLenL:
            if parent != ROOT_PARENT_NAME:
                branchLenD[(parent,child)] = brLen
        if all(brLen==None for brLen in branchLenD.values()):
            # if all branches None, don't bother with dict
            branchLenD = None
        
        # put in these attributes
        self.populateAttributes(nodeConnectD,arbitraryNode,branchLenD)

    def toNewickStr(self,includeBrLength=False,nodeLabelD=None):
        '''Output a newick string.'''
        if self.nodeCount() == 1:
            return "("+self.leafNodeT[0]+")"
        elif self.nodeCount() == 2:
            # two tip tree is a special case            
            return "("+self.leafNodeT[0]+","+self.leafNodeT[1]+")"
        else:
            return self.__traverseForNewickStr__(self.arbitraryNode,ROOT_PARENT_NAME,includeBrLength,nodeLabelD)

    def root(self,branchToRootPair):
        '''Root using the tuple branchToRootPair and return an Rtree object.'''

        def updateChildOfRoot(newD,nodeToWorkOn,nodeToReplace):
            '''Adjust child of root to say root is parent.'''
            newConnecL = ["root"]
            for node in newD[nodeToWorkOn]:
                if node != nodeToReplace:
                    newConnecL.append(node)
            newD[nodeToWorkOn] = tuple(newConnecL)
            return newD
            
        # make new nodeConnectD
        newD = {}
        self.__splitNodeConnectD__(self.nodeConnectD,newD,branchToRootPair[0],branchToRootPair[1])
        self.__splitNodeConnectD__(self.nodeConnectD,newD,branchToRootPair[1],branchToRootPair[0])

        # must add the root
        rootNode = "root"
        newD[rootNode] = (ROOT_PARENT_NAME,branchToRootPair[0],branchToRootPair[1])

        # adjust children of root to say root is parent
        newD = updateChildOfRoot(newD,branchToRootPair[0],branchToRootPair[1])
        newD = updateChildOfRoot(newD,branchToRootPair[1],branchToRootPair[0])
        
        return Rtree(newD,rootNode)

    def rootIncludeBranchLen(self,branchToRootPair):
        '''Root using the tuple branchToRootPair and return an Rtree object that
includes branch lengths.'''

        # output tree
        rtreeO = self.root(branchToRootPair)

        # get all branch lengths besides those involving the root
        branchLenD = {}
        for utreeBranchPair in self.branchPairT:

            if utreeBranchPair == branchToRootPair or utreeBranchPair == (branchToRootPair[1],branchToRootPair[0]):
                # this is the branch to be rooted, skip
                continue
            else:
                # make sure its in preorder as defined for rtreeO
                rtreeBranchPair = tuple((nd for nd in rtreeO.preorder() if nd in utreeBranchPair))

                branchLenD[rtreeBranchPair] = self.branchLenD[utreeBranchPair]

        # now get branch lens involving root. Arbitrarily split 50:50.
        utreeBranchToRootLen = self.branchLenD[branchToRootPair]
        branchLenD[("root",branchToRootPair[0])] = 0.5 * utreeBranchToRootLen
        branchLenD[("root",branchToRootPair[1])] = 0.5 * utreeBranchToRootLen
        
        # add to tree
        rtreeO.branchLenD = branchLenD

        return rtreeO
        
    def iterAllRootedTrees(self):
        '''Iterator yielding all possible rooted trees from this unrooted tree.'''
        for branchPair in self.branchPairT:
            yield self.root(branchPair)

    def iterAllRootedTreesIncludeBranchLen(self):
        '''Iterator yielding all possible rooted trees from this unrooted tree.'''
        for branchPair in self.branchPairT:
            yield self.rootIncludeBranchLen(branchPair)

    def split(self,branchPair):
        '''Split on the branch specified by branchPair into two new Utree
objects.

        '''
        
        def subUtree(oldNodeConnectD,oldBranchPairT,oldBranchLenD,node,parentNode):
            '''Given an oldNodeConnectD from a Utree object, and a node and it's
parent, create a new Utree object which is a sub tree. Assumes node is
not a tip.

            '''
            # get subset of oldNodeConnectD, put in D
            D = {}
            self.__splitNodeConnectD__(oldNodeConnectD,D,node,parentNode)

            newBranchLenD = {}
            if len(D[node])>3:
                # it's a multifurcating node, node should remain in D,
                # but we must remove the connection to parentNode
                connecT = D[node]
                newConnecL = []
                for tempNode in connecT:
                    if tempNode != parentNode:
                        newConnecL.append(tempNode)
                D[node] = tuple(newConnecL)

                # create output tree
                newUtreeO = Utree(D)
                # get branch lengths
                for nbp in newUtreeO.branchPairT:
                    # might need to reverse order of nbp to find in oldBranchLenD
                    if nbp in oldBranchLenD:
                        newBranchLenD[nbp] = oldBranchLenD[nbp]
                    else:
                        newBranchLenD[nbp] = oldBranchLenD[(nbp[1],nbp[0])]
                   
            else:
                # bifurcating node
                # get the two branches that connect to node so we can
                # remove node itself.
                _,child1,child2 = D[node]

                assert(_==parentNode) # temp

                # adjust entries for children
                updateChild(D,node,child1,child2)
                updateChild(D,node,child2,child1)
                del D[node]
            
                # create output tree
                newUtreeO = Utree(D)

                # get branch lengths
                for nbp in newUtreeO.branchPairT:
                    if child1 in nbp and child2 in nbp:
                        # this branch len must be made by adding two old ones
                        oldBP1 = [oldBP for oldBP in oldBranchPairT if (node in oldBP and child1 in oldBP)][0]
                        oldBP2 = [oldBP for oldBP in oldBranchPairT if (node in oldBP and child2 in oldBP)][0]
                        brLen = oldBranchLenD[oldBP1] + oldBranchLenD[oldBP2]
                        newBranchLenD[nbp] = brLen
                    else:
                        # might need to reverse order of nbp to find in oldBranchLenD
                        if nbp in oldBranchLenD:
                            newBranchLenD[nbp] = oldBranchLenD[nbp]
                        else:
                            newBranchLenD[nbp] = oldBranchLenD[(nbp[1],nbp[0])]
                        
            newUtreeO.branchLenD = newBranchLenD
            return newUtreeO

        def updateChild(D,node,child1,child2):
            '''Update entries in D for child1 to connect to child2 rather than
node.'''
            connecL = list(D[child1])
            connecL[connecL.index(node)] = child2
            D[child1] = tuple(connecL)
            return

        # main part of split
        if self.isLeaf(branchPair[0]) and self.isLeaf(branchPair[1]):
            return Utree({branchPair[0]:()}),Utree({branchPair[1]:()}) # branchLenD None by default
        elif self.isLeaf(branchPair[0]):
            # one side is leaf
            restUtreeO = subUtree(self.nodeConnectD,self.branchPairT,self.branchLenD,branchPair[1],branchPair[0])
            return Utree({branchPair[0]:()}),restUtreeO
        elif self.isLeaf(branchPair[1]):
            restUtreeO = subUtree(self.nodeConnectD,self.branchPairT,self.branchLenD,branchPair[0],branchPair[1])
            return restUtreeO,Utree({branchPair[1]:()})
        else:
            # internal
            aUtreeO = subUtree(self.nodeConnectD,self.branchPairT,self.branchLenD,branchPair[1],branchPair[0])
            bUtreeO = subUtree(self.nodeConnectD,self.branchPairT,self.branchLenD,branchPair[0],branchPair[1])
            return aUtreeO,bUtreeO
        
    def __bioPhyloToNodeConnectD__(self,bpClade,parentNodeStr,nodeConnectD,iNodeNum):
        '''Convert a biopython clade object to a node connection dict with
keys that are nodes, and values that are the nodes connected to. For
internals (node1, node2, node3...). We assume bpClade does not have
named internal nodes, and we name them here: g0, g1 etc.

        '''

        if bpClade.is_terminal():
            # parentNodeStr will never be ROOT_PARENT_NAME in this case
            nodeConnectD[bpClade.name] = (parentNodeStr,)
            branchLenL = [(parentNodeStr,bpClade.name,bpClade.branch_length)]

        elif parentNodeStr == ROOT_PARENT_NAME and len(bpClade) == 2:
            # biopython has put 2 branches only at the base. We will
            # operate over this branch, joining the two nodes with no
            # node in between.

            if bpClade[0].is_terminal() and bpClade[1].is_terminal():
                # both terminal
                iNodeNum,nodeConnectD,_ = self.__bioPhyloToNodeConnectD__(bpClade[0],bpClade[1].name,nodeConnectD,iNodeNum)
                iNodeNum,nodeConnectD,_ = self.__bioPhyloToNodeConnectD__(bpClade[1],bpClade[0].name,nodeConnectD,iNodeNum)
                
                # get branches
                if bpClade[0].branch_length == None or bpClade[1].branch_length == None:
                    brLen = None
                else:
                    brLen = bpClade[0].branch_length + bpClade[1].branch_length
                branchLenL = [(bpClade[0].name,bpClade[1].name,brLen)]
                
            elif bpClade[0].is_terminal():
                # 0 is terminal, 1 not
                # parent node for 0 will be based on current iNodeNum
                iNodeNum,nodeConnectD,_ = self.__bioPhyloToNodeConnectD__(bpClade[0],"g"+str(iNodeNum),nodeConnectD,iNodeNum)
                iNodeNum,nodeConnectD,branchLenL = self.__bioPhyloToNodeConnectD__(bpClade[1],bpClade[0].name,nodeConnectD,iNodeNum)

            elif bpClade[1].is_terminal():
                # 1 is terminal, 0 not
                # parent node for 1 will be based on current iNodeNum
                iNodeNum,nodeConnectD,_ = self.__bioPhyloToNodeConnectD__(bpClade[1],"g"+str(iNodeNum),nodeConnectD,iNodeNum)
                iNodeNum,nodeConnectD,branchLenL = self.__bioPhyloToNodeConnectD__(bpClade[0],bpClade[1].name,nodeConnectD,iNodeNum)

            else:
                # neither is terminal

                # get the iNodeNum for the first internal node in 0
                iNodeNumAtBaseOf0 = iNodeNum
                
                # figure out what the parent for 0 will be by making a
                # dummy call. iNodeNum afterward will be the first
                # internal node in 1.
                iNodeNumAtBaseOf1,_,_ = self.__bioPhyloToNodeConnectD__(bpClade[0],"dummy",{},iNodeNum)

                # now make real calls
                branchLenL = []
                iNodeNum,nodeConnectD,tempBranchLenL = self.__bioPhyloToNodeConnectD__(bpClade[0],"g"+str(iNodeNumAtBaseOf1),nodeConnectD,iNodeNum)
                branchLenL.extend(tempBranchLenL)

                iNodeNum,nodeConnectD,tempBranchLenL = self.__bioPhyloToNodeConnectD__(bpClade[1],"g"+str(iNodeNumAtBaseOf0),nodeConnectD,iNodeNum)

                # there is a redundant entry in tempBranchLenL. Don't
                # include in branchLenL. The entry that has
                # ("g"+str(iNodeNumAtBaseOf0),"g"+str(iNodeNumAtBaseOf1),brLen)
                for brLenT in tempBranchLenL:
                    if brLenT[0] != "g"+str(iNodeNumAtBaseOf0) or brLenT[1] != "g"+str(iNodeNumAtBaseOf1):
                        branchLenL.append(brLenT)
        else:
            # general case
            connecL = []
            branchLenL = []
            if parentNodeStr != ROOT_PARENT_NAME:
                connecL.append(parentNodeStr)
            # if we're just starting, it's ROOT_PARENT_NAME, and we
            # don't want to put that in the list of connections
            thisNodeStr = "g"+str(iNodeNum)
            iNodeNum += 1
            for childClade in bpClade:

                # collect connection
                if childClade.is_terminal():
                    connecL.append(childClade.name)
                else:
                    # the connecting node will be the next one (iNodeNum)
                    connecL.append("g"+str(iNodeNum))

                # recurse
                iNodeNum,nodeConnectD,tempBranchLenL = self.__bioPhyloToNodeConnectD__(childClade,thisNodeStr,nodeConnectD,iNodeNum)
                branchLenL.extend(tempBranchLenL)
                
            # add in this connection for this node
            nodeConnectD[thisNodeStr] = tuple(connecL)

            # get branch len
            if hasattr(bpClade,'branch_length'):
                brLen = bpClade.branch_length
            else:
                brLen = None
            branchLenL.append((parentNodeStr,thisNodeStr,brLen))

            
        return iNodeNum,nodeConnectD,branchLenL
    
    def __splitNodeConnectD__(self,D,newD,node,parentNode):
        '''Get nodeConnectD for the part of the tree defined by node, and in
the oposite direction from parentNode. Put in newD. Returns None.

        '''
        oldConnecT=D[node]

        # make sure parent node is first
        assert(parentNode in oldConnecT)
        connecL = [parentNode]
        for tempNode in oldConnecT:
            if tempNode != parentNode:
                connecL.append(tempNode)
        connecT = tuple(connecL)

        newD[node] = connecT # store

        if len(connecT)==1:
            return
        else:
            for child in connecT:
                if child != parentNode:
                    self.__splitNodeConnectD__(D,newD,child,node)
            return
        
    def __repr__(self):
        return "Utree: "+self.toNewickStr()



# Greedy.py
#Ivy Liu  
# Modified from code by
# Srinidhi Srinivasan, Juliet Forman
# June 2015

# This file contains the functions for finding all optimal reconciliations 
# using a DTL graph, as well as the scores for each of the events using the
# vertex-based DP algorithm. The main function in this file is called Greedy
# and the remaining functions are helper functions that are used by Greedy.

import copy 


def findRoot(Tree):
    """This function takes in a parasiteTree and returns a string with the 
    name of the root vertex of the tree"""

    if 'pTop' in Tree:
        return Tree['pTop'][1]
    return Tree['hTop'][1] 


def orderDTLOR(DTLOR, ParasiteRoot):
    """This function takes in a DTLOR graph and the ParasiteRoot. It outputs a 
    list, keysL, that contains tuples. Each tuple has two elements. The first
    is a mapping node of the form (p, h, l, l), where p is a parasite node and h is 
    a host node, l is the syntenic location of p. The second element is a level representing the depth of that
    mapping node within the tree."""

    keysL = []
    topNodes = []
    for key in DTLOR:
        if key[0] == ParasiteRoot:
            topNodes.append(key)
    for vertex in topNodes:
        keysL.extend(orderDTLOR_Roots(DTLOR, vertex, 0))
    return keysL


def orderDTLOR_Roots(DTLOR, vertex, level):
    """This function takes a DTL graph, one node, a vertex, of the DTL graph, 
    and level, and returns a list, keysL, that contains tuples. Each tuple has
    two elements. The first is a mapping node of the form (p, h, l, l), where p is a
    parasite node, h is a host node and l is the syntenic location of p. 
    The second element is a level representing the depth of that mapping node 
    within the tree. This function adds the input vertex to keysL and recurses 
    on its children."""

    keysL = []
    # Loop through each event associated with key in DTL
    for i in range(len(DTLOR[vertex]) - 1):
        event = DTLOR[vertex][i]
        child1 = event[1]
        child2 = event[2]
        keysL = keysL + [(vertex, level)]
        if child1[0] is not None:
            keysL.extend(orderDTLOR_Roots(DTLOR, child1, level + 1))
        if child2[0] is not None:
            keysL.extend(orderDTLOR_Roots(DTLOR, child2, level + 1)) 
    return keysL


def sortHelper(DTLOR, keysL):
    """This function takes in a list keysL,and a DTLOR graph, and returns a new 
    list, uniqueKeysL that has removed duplicate mapping nodes from keysL.
    This function chooses the highest level for each mapping node because we 
    are using the bottom-up approach."""
    
    uniqueKeysL = []
    for key in DTLOR:
        maxLevel = float("-inf")
        for element in keysL:
            if key == element[0]:
                if element[1] > maxLevel:
                    maxLevel = element[1]
        uniqueKeysL.append((key, maxLevel))
    return uniqueKeysL


def postorderDTLsort(DTLOR, ParasiteRoot):
    """This function takes in a DTL graph and ParasiteRoot, and returns a 
    sorted list, orderedKeysL, that is ordered by level from largest to 
    smallest, where level 0 is the root and the highest level are tips."""

    keysL = orderDTLOR(DTLOR, ParasiteRoot)
    uniqueKeysL = sortHelper(DTLOR, keysL)
    orderedKeysL = []
    levelCounter = 0
    while len(orderedKeysL) < len(uniqueKeysL):
        for mapping in uniqueKeysL:
            if mapping[-1] == levelCounter:
                orderedKeysL = [mapping] + orderedKeysL
        levelCounter += 1
    return orderedKeysL


def bookkeeping(DTLOR, ParasiteTree):
    """This function takes as inputs a DTLOR graph and ParasiteTree, and then 
    records what the max is at each mapping node and which event the max came 
    from. It outputs a dictionary BSFHMap by looping through the keys in a 
    sorted list of mapping nodes and finding the max score at each mapping 
    node and event node. BSFHMap has keys of the form (p, h, l, l) which are the 
    mapping nodes, and values which are lists where the first element is an 
    event with the max score, and the last element is the maxScore."""

    # Example: BSFHMap = {(mapping node): [['event', (p, h, l, l), (p, h, l, l), score], maxScore]}
    # Example: BSFHEvent = {(event node): max}

    BSFHMap = {}
    BSFHEvent = {}
    ParasiteRoot = findRoot(ParasiteTree)
    orderedKeysL = postorderDTLsort(DTLOR, ParasiteRoot)   
    for key in orderedKeysL:
        mapNode = key[0]
        # Check if the key is a tip:
        if DTLOR[mapNode][0][0] == 'C':              
            BSFHMap[mapNode] = [DTLOR[mapNode][0], DTLOR[mapNode][0][-1]]
        # If the key isn't a tip:
        else: 
            # Initialize counter
            maxScore = float("-inf")  
            # Initialize variable to keep track of where max came from
            maxEvent = [] 
            # Iterate through the events associated with the key node
            for i in range(len(DTLOR[mapNode]) - 1):  
                event = tuple(DTLOR[mapNode][i])
                BSFHEvent[event] = event[-1]
                if event[1] != (None, None, None, None):
                    BSFHEvent[event] = BSFHEvent[event]+BSFHMap[event[1]][-1]
                if event[2] != (None, None, None, None):
                    BSFHEvent[event] = BSFHEvent[event]+BSFHMap[event[2]][-1]
                # Check if current event has a higher score than current max
                if BSFHEvent[event] > maxScore:  
                    maxScore = BSFHEvent[event]  # Set new max score
                    maxEvent = list(event)   # Record where new max came from
            BSFHMap[mapNode] = [maxEvent, maxScore]  # Set BSFH value of key
    return BSFHMap


def TraceChildren(DTLOR, GreedyOnce, BSFHMap, key):
    """This function takes as input a DTLOR graph, a dictionary GreedyOnce, 
    containing the root of an optimal reconciliation, a BSFHMap dictionary,
    and a current key, and returns the optimal reconciliation and new DTLOR 
    graph, where the scores of the events in the reconciliation are reset to
    0."""

    resetDTLOR = {}  # The new DTL graph
    reset1DTLOR = {}  # This DTL graph deals with the recursive call on the child1
    reset2DTLOR = {}  # This DTL graph deals with the recursive call on the child2
    child1 = GreedyOnce[key][1]
    child2 = GreedyOnce[key][2]
    if child1 != (None, None, None, None):
        GreedyOnce[child1] = BSFHMap[child1][0][0:3]  # Add event to greedyOnce, without the score
        # This loop resets all the scores of events that have been used to 0
        for i in range(len(DTLOR[child1]) - 1):       
            if DTLOR[child1][i] == BSFHMap[child1][0]:
                newValue = DTLOR[child1]
                newValue[i][-1] = 0
                reset1DTLOR[child1] = newValue
        # This recursive call updates GreedyOnce and the DTL graph
        newGreedyOnce, DTLOR1 = TraceChildren(DTLOR, GreedyOnce, BSFHMap, child1) 
        reset1DTLOR.update(DTLOR1)
        GreedyOnce.update(newGreedyOnce)
    if child2 != (None, None, None, None):
        GreedyOnce[child2] = BSFHMap[child2][0][0:3]
        for i in range(len(DTLOR[child2]) - 1): 
            if DTLOR[child2][i] == BSFHMap[child2][0]:
                newValue = DTLOR[child2]
                newValue[i][-1] = 0
                reset2DTLOR[child2] = newValue      
        newGreedyOnce, DTLOR2 = TraceChildren(DTLOR, GreedyOnce, BSFHMap, child2)
        reset2DTLOR.update(DTLOR2)
        GreedyOnce.update(newGreedyOnce)
    resetDTLOR.update(reset1DTLOR)
    resetDTLOR.update(reset2DTLOR)
    return GreedyOnce, resetDTLOR


def greedyOnce(DTLOR, ParasiteTree):
    """This function takes the DTL graph and the ParasiteTree as inputs, and 
    calls bookkeeping to find the dictionary BSFHMap. It returns the 
    reconciliation with the highest score in a dictionary called GreedyOnce, 
    and also resets to 0 the scores of the mapping nodes in the optimal 
    reconciliation. The returned dictionary will have keys which are the 
    mapping nodes in the optimal reconciliation, and values of the form 
    (event, child1, child2)."""

    BSFHMap = bookkeeping(DTLOR, ParasiteTree)
    ParasiteRoot = findRoot(ParasiteTree)
    GreedyOnce = {}            # Initialize dictionary we will return
    bestKey = ()       # Variable to hold the key with the highers BSFH value
    bestScore = float("-inf")  # Variable to hold the highest BSFH value so far
    # Iterate trough all the keys (vertices) in BSFHMap
    for key in BSFHMap:   
        # Check if key has a score higher than bestScore and includes PRoot
        if BSFHMap[key][-1] > bestScore and key[0] == ParasiteRoot: 
            bestKey = key
            bestScore = BSFHMap[key][-1]
    # Set value in GreedyOnce of the best key we found
    GreedyOnce[bestKey] = BSFHMap[bestKey][0][0:3]                  
    
    # Reset score of the mapping node we used to 0:

    # Loop through the events associated with DTL
    for i in range(len(DTLOR[bestKey]) - 1):  
        # Check if the event matches the event that gave the best score
        if DTLOR[bestKey][i] == BSFHMap[bestKey][0]:           
            newEvent = DTLOR[bestKey][i][:-1] + [0]
            newValue = DTLOR[bestKey][:i] + [newEvent] + DTLOR[bestKey][i + 1:]
            DTLOR[bestKey] = newValue  # Set the score to 0
    newGreedyOnce, resetDTLOR = TraceChildren(DTLOR, GreedyOnce, BSFHMap, bestKey)
    GreedyOnce.update(newGreedyOnce)
    DTLOR.update(resetDTLOR)
    return GreedyOnce, DTLOR, bestScore


def Greedy(DTLOR, ParasiteTree):
    """This function takes as input a DTL graph and a ParasiteTree, and 
    returns TreeList, a list of dictionaries, each of which represent one of 
    the optimal reconciliations. This function runs till all the scores have 
    been collected from the DTL graph."""
    scores = []  # List of reconciliation scores
    currentDTLOR = copy.deepcopy(DTLOR)
    rec = [] # List of reconciliations
    collected = True
    while collected:
        # Call greedyOnce if all the points have not been collected yet
        oneTree, currentDTLOR, score = greedyOnce(currentDTLOR, ParasiteTree)
        scores.append(score) 
        rec.append(oneTree)
        collected = False  # Set not0 to False
        # Iterate to see if more points need to be collected
        for key in currentDTLOR:
            for i in range(len(currentDTLOR[key])-1):
                if currentDTLOR[key][i][-1] != 0:
                    collected = True
    return scores, rec
